
from collections import defaultdict


from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.utils import unique_ever_seen

from ..equivariant_module import EquivariantModule

import torch
from torch.nn import Parameter
from typing import List, Tuple, Any, Union
import numpy as np

__all__ = ["GNormBatchNorm"]


class GNormBatchNorm(EquivariantModule):

    def __init__(self,
                 in_type: FieldType,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,
                 track_running_stats: bool = True,
                 ):
        r"""

        Batch normalization for generic representations.

        This batch normalization assumes that the covariance matrix of a subset of channels in `in_type` transforming
        under an irreducible representation is a scalar multiple of the identity.
        Moreover, the mean is only computed over the trivial irreps occurring in the input representations.
        These assumptions are necessary and sufficient conditions for the equivariance in expectation of this module,
        see Chapter 4.2 at `https://gabri95.github.io/Thesis/thesis.pdf <https://gabri95.github.io/Thesis/thesis.pdf>`_ .

        Similarly, if ``affine = True``, a single scale is learnt per input irrep and the bias is applied only to the
        trivial irreps.

        Note that the representations in the input field type do not need to be already decomposed into direct sums of
        irreps since this module can deal with changes of basis.

        .. warning::
            However, because the irreps in the input representations rarely appear in a contiguous way, this module might
            internally use advanced indexing, leading to some computational overhead.
            Modules like :class:`~escnn.nn.IIDBatchNorm2d` or :class:`~escnn.nn.IIDBatchNorm3d`, instead, share the same
            variance with all channels within the same field (and, therefore, over multiple irreps).
            This can be more efficient if the input field type contains multiple copies of a larger,
            reducible representation.


        Args:
            in_type (FieldType): the input field type
            eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
            momentum (float, optional): the value used for the ``running_mean`` and ``running_var`` computation.
                    Can be set to ``None`` for cumulative moving average (i.e. simple average). Default: ``0.1``
            affine (bool, optional): if ``True``, this module has learnable affine parameters. Default: ``True``
            track_running_stats (bool, optional): when set to ``True``, the module tracks the running mean and variance;
                                                  when set to ``False``, it does not track such statistics but uses
                                                  batch statistics in both training and eval modes.
                                                  Default: ``True``

        """

        assert isinstance(in_type.gspace, GSpace)
        
        super(GNormBatchNorm, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self.affine = affine
        self.track_running_stats = track_running_stats

        self._nfields = None
        
        # group fields by their type and
        #   - check if fields of the same type are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each type
        self._nfields = defaultdict(int)
        
        # indices of the channels corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        ntrivials = 0
        position = 0
        last_field = None
        for i, r in enumerate(self.in_type.representations):
            
            for irr in r.irreps:
                if self.in_type.fibergroup._irreps[irr].is_trivial():
                    ntrivials += 1
            
            if r.name != last_field:
                if not r.name in self._contiguous:
                    self._contiguous[r.name] = True
                else:
                    self._contiguous[r.name] = False

            last_field = r.name
            _indices[r.name] += list(range(position, position + r.size))
            self._nfields[r.name] += 1
            position += r.size
        
        for name, contiguous in self._contiguous.items():
            
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[name] = [min(_indices[name]), max(_indices[name])+1]
                setattr(self, f"{self._escape_name(name)}_indices", _indices[name])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[name] = torch.LongTensor(_indices[name])
                
                # register the indices tensors as parameters of this module
                self.register_buffer(f"{self._escape_name(name)}_indices", _indices[name])
        
        # store the size of each field type
        self._sizes = []
        
        # store for each field type the indices of the trivial irreps in it
        self._trivial_idxs = {}

        # store for each field type the sizes and the indices of all its irreps, grouped by their size
        self._irreps_sizes = {}

        for r in unique_ever_seen(self.in_type.representations):
            p = 0
            irreps = defaultdict(lambda: [])
            trivials = []
            aggregator = torch.zeros(r.size, len(r.irreps))
            
            for i, irr in enumerate(r.irreps):
                irr = self.in_type.fibergroup._irreps[irr]
                if irr.is_trivial():
                    trivials.append(p)
                
                aggregator[p:p+irr.size, i] = 1. / irr.size
                
                irreps[irr.size] += list(range(p, p+irr.size))
                p += irr.size
            
            propagator = (aggregator > 0).clone().to(dtype=torch.float)
            
            name = r.name
            
            self._trivial_idxs[name] = torch.tensor(trivials, dtype=torch.long)
            self._irreps_sizes[name] = [(s, idxs) for s, idxs in irreps.items()]
            self._sizes.append((name, r.size))
            
            if not np.allclose(r.change_of_basis, np.eye(r.size)):
                self.register_buffer(f'{self._escape_name(name)}_change_of_basis', torch.tensor(r.change_of_basis, dtype=torch.float))
                self.register_buffer(f'{self._escape_name(name)}_change_of_basis_inv', torch.tensor(r.change_of_basis_inv, dtype=torch.float))
            
            self.register_buffer(f'vars_aggregator_{self._escape_name(name)}', aggregator)
            self.register_buffer(f'vars_propagator_{self._escape_name(name)}', propagator)
        
            running_var = torch.ones((self._nfields[r.name], len(r.irreps)), dtype=torch.float)
            running_mean = torch.zeros((self._nfields[r.name], len(trivials)), dtype=torch.float)
            self.register_buffer(f'{self._escape_name(name)}_running_var', running_var)
            self.register_buffer(f'{self._escape_name(name)}_running_mean', running_mean)
            
            if self.affine:
                weight = Parameter(torch.ones((self._nfields[r.name], len(r.irreps))), requires_grad=True)
                bias = Parameter(torch.zeros((self._nfields[r.name], len(trivials))), requires_grad=True)
                self.register_parameter(f'{self._escape_name(name)}_weight', weight)
                self.register_parameter(f'{self._escape_name(name)}_bias', bias)
            
        self.register_buffer('num_batches_tracked', torch.tensor(0, dtype=torch.long))
        
        self.eps = eps
        self.momentum = momentum

    def reset_running_stats(self):
        for name, size in self._sizes:
            running_var = getattr(self, f"{self._escape_name(name)}_running_var")
            running_mean = getattr(self, f"{self._escape_name(name)}_running_mean")
            running_var.fill_(1)
            running_mean.fill_(0)
        self.num_batches_tracked.zero_()

    def reset_parameters(self):
        self.reset_running_stats()
        if self.affine:
            for name, size in self._sizes:
                weight = getattr(self, f"{self._escape_name(name)}_weight")
                bias = getattr(self, f"{self._escape_name(name)}_bias")
                weight.data.fill_(1)
                bias.data.fill_(0)

    def _get_running_stats(self, name: str):
        vars = getattr(self, f"{self._escape_name(name)}_running_var")
        means = getattr(self, f"{self._escape_name(name)}_running_mean")
        return means, vars

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        Apply norm non-linearities to the input feature map
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type

        exponential_average_factor = 0.0

        if self.training:
            self.num_batches_tracked += 1
            if self.momentum is None:  # use cumulative moving average
                exponential_average_factor = 1.0 / self.num_batches_tracked.item()
            else:  # use exponential moving average
                exponential_average_factor = self.momentum

        coords = input.coords
        input = input.tensor
        shape = input.shape[2:]
        b, c = input.shape[:2]
        
        output = torch.empty_like(input)
        
        # iterate through all field types
        for name, size in self._sizes:
            indices = getattr(self, f"{self._escape_name(name)}_indices")
            
            if self._contiguous[name]:
                slice = input[:, indices[0]:indices[1], ...]
            else:
                slice = input[:, indices, ...]

            slice = slice.view(b, -1, size, *shape)
            
            if hasattr(self, f"{self._escape_name(name)}_change_of_basis_inv"):
                cob_inv = getattr(self, f"{self._escape_name(name)}_change_of_basis_inv")
                slice = torch.einsum("ds,bcsxy->bcdxy", (cob_inv, slice))

            if not self.track_running_stats:
                means, vars = self._compute_statistics(slice, name)
            elif self.training:
                
                # compute the mean and variance of the fields
                means, vars = self._compute_statistics(slice, name)
                
                running_mean, running_var = self._get_running_stats(name)

                running_var *= 1 - exponential_average_factor
                running_var += exponential_average_factor * vars
                
                running_mean *= 1 - exponential_average_factor
                running_mean += exponential_average_factor * means

                assert not torch.isnan(running_mean).any()
                assert not torch.isnan(running_var).any()
                assert torch.allclose(running_mean, getattr(self, f"{self._escape_name(name)}_running_mean"))
                assert torch.allclose(running_var, getattr(self, f"{self._escape_name(name)}_running_var"))
                
            else:
                means, vars = self._get_running_stats(name)

            if self.affine:
                weight = getattr(self, f"{self._escape_name(name)}_weight")
            else:
                weight = 1.

            # helps in case mixed precision is used
            vars[vars < 0.] = 0.

            # compute the scalar multipliers needed
            scales = weight / (vars + self.eps).sqrt()

            # compute the point shifts
            # shifts = bias - self._scale(means, scales, name=name)
            centered = self._shift(slice, -1*means, name=name, out=None)
            normalized = self._scale(centered, scales, name=name, out=None)
            
            if self.affine:
                bias = getattr(self, f"{self._escape_name(name)}_bias")
                normalized = self._shift(normalized, bias, name=name, out=None)
            
            if hasattr(self, f"{self._escape_name(name)}_change_of_basis"):
                cob = getattr(self, f"{self._escape_name(name)}_change_of_basis")
                normalized = torch.einsum("ds,bcsxy->bcdxy", (cob, normalized))
                
            if not self._contiguous[name]:
                output[:, indices, ...] = normalized.view(b, -1, *shape)
            else:
                output[:, indices[0]:indices[1], ...] = normalized.view(b, -1, *shape)

            # if self._contiguous[name]:
            #     slice2 = output[:, indices[0]:indices[1], ...]
            # else:
            #     slice2 = output[:, indices, ...]
            # assert torch.allclose(slice2.view(b, -1, size, h, w), slice), name
            
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(NormBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass

    def _compute_statistics(self, t: torch.Tensor, name: str):
        
        trivial_idxs = self._trivial_idxs[name]
        vars_aggregator = getattr(self, f"vars_aggregator_{self._escape_name(name)}")
        
        b, c, s = t.shape[:3]
        shape = t.shape[3:]

        l = trivial_idxs.numel()
        
        # number of samples in the tensor used to estimate the statistics
        M = (np.prod(shape) if shape else 1)
        N = b * M

        # compute the mean of the trivial fields
        trivial_means = t[:, :, trivial_idxs, ...].view(b, c, l, M).mean(dim=(0, 3), keepdim=False).detach()
        
        # compute the mean of squares of all channels
        vars = (t ** 2).view(b, c, s, M).mean(dim=(0, 3), keepdim=False).detach()
        
        # For the non-trivial fields the mean of the fields is 0, so we can compute the variance as the mean of the
        # norms squared.
        # For trivial channels, we need to subtract the squared mean
        vars[:, trivial_idxs] -= trivial_means**2

        # helps in case mixed precision is used
        vars[vars < 0.] = 0.

        # aggregate the squared means of the channels which belong to the same irrep
        vars = torch.einsum("io,ci->co", (vars_aggregator, vars))

        # helps in case mixed precision is used
        vars[vars < 0.] = 0.

        # Correct the estimation of the variance with Bessel's correction
        correction = N/(N-1) if N > 1 else 1.
        vars *= correction
        
        return trivial_means, vars

    def _scale(self, t: torch.Tensor, scales: torch.Tensor, name: str, out: torch.Tensor = None):
        
        if out is None:
            out = torch.empty_like(t)
        
        vars_aggregator = getattr(self, f"vars_propagator_{self._escape_name(name)}")
        
        ndims = len(t.shape[3:])
        scale_shape = (1, scales.shape[0], vars_aggregator.shape[0]) + (1,)*ndims
        # scale all fields
        out[...] = t * torch.einsum("oi,ci->co", (vars_aggregator, scales)).reshape(scale_shape)
        
        return out
    
    def _shift(self, t: torch.Tensor, trivial_bias: torch.Tensor, name: str, out: torch.Tensor = None):
    
        if out is None:
            out = t.clone()
        else:
            out[:] = t
            
        trivial_idxs = self._trivial_idxs[name]
        
        bias_shape = (1,) + trivial_bias.shape + (1,)*(len(t.shape) - 3)
        
        # add bias to the trivial fields
        out[:, :, trivial_idxs, ...] += trivial_bias.view(bias_shape)

        return out

    def _escape_name(self, name: str):
        return name.replace('.', '^')

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')

        main_str = self._get_name() + '('
        if len(extra_lines) == 1:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(extra_lines) + '\n'

        main_str += ')'
        return main_str

    def extra_repr(self):
        return '{in_type}, eps={eps}, momentum={momentum}, affine={affine}, track_running_stats={track_running_stats}' \
            .format(**self.__dict__)

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.BatchNormNd` module and set to "eval" mode.

        """

        if not self.track_running_stats:
            raise ValueError('''
                Equivariant Batch Normalization can not be converted into conventional batch normalization when
                "track_running_stats" is False because the statistics contained in a single batch are generally
                not symmetric
            ''')

        self.eval()

        batchnorm = torch.nn.BatchNorm3d(
            self.in_type.size,
            self.eps,
            self.momentum,
            affine=self.affine,
            track_running_stats=self.track_running_stats
        )

        batchnorm.num_batches_tracked.data = self.num_batches_tracked.data

        for name, size in self._sizes:
            contiguous = self._contiguous[name]
            if not contiguous:
                raise NotImplementedError(
                    '''Non-contiguous indices not supported yet when converting
                    inner-batch normalization into conventional BatchNorm2d'''
                )

            n = self._nfields[name]

            start, end = getattr(self, f"{self._escape_name(name)}_indices")

            running_mean, running_var = self._get_running_stats(name)

            batchnorm.running_mean.data[start:end] = self._shift(torch.zeros(
                1, n, size
            ), running_mean, name=name, out=None).view(-1)
            batchnorm.running_var.data[start:end] = self._scale(torch.ones(
                1, n, size
            ), running_var, name=name, out=None).view(-1)

            if self.affine:
                weight = getattr(self, f'{self._escape_name(name)}_weight')
                batchnorm.weight.data[start:end] = self._scale(torch.ones(
                    1, n, size
                ), weight, name=name, out=None).view(-1)

                bias = getattr(self, f"{self._escape_name(name)}_bias")
                batchnorm.bias.data[start:end] = self._shift(torch.zeros(
                    1, n, size
                ), bias, name=name, out=None).view(-1)

            if hasattr(self, f"{self._escape_name(name)}_change_of_basis"):
                cob = getattr(self, f"{self._escape_name(name)}_change_of_basis")

                batchnorm.running_var.data[start:end] = torch.einsum(
                    "ds,cs->cd",
                    cob,
                    batchnorm.running_var.data[start:end].view(n, size)
                ).view(-1)

        batchnorm.eval()

        return batchnorm







from escnn.nn import FieldType

from .pointwise import _PointwisePoolND, _PointwiseAvgPoolAntialiasedND

import torch

from typing import Tuple, Union, Optional

__all__ = [
    "PointwiseAvgPool",
    "PointwiseAvgPoolAntialiased",
    "PointwiseAvgPool2D",
    "PointwiseAvgPoolAntialiased2D",
]


class PointwiseAvgPool2D(_PointwisePoolND):
    
    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Optional[Union[int, Tuple[int, int]]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise average-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AvgPool2d`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a average over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape

        """
        super().__init__(
            in_type, 2,
            pool=torch.nn.AvgPool2d(
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                ceil_mode=ceil_mode,
            ),
        )
    
    
class PointwiseAvgPool3D(_PointwisePoolND):
    
    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int, int]],
                 stride: Optional[Union[int, Tuple[int, int, int]]] = None,
                 padding: Union[int, Tuple[int, int, int]] = 0,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise average-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AvgPool3d`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a average over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape

        """
        super().__init__(
            in_type, 3,
            pool=torch.nn.AvgPool3d(
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                ceil_mode=ceil_mode,
            ),
        )
    
    
class PointwiseAdaptiveAvgPool2D(_PointwisePoolND):

    def __init__(self,
                 in_type: FieldType,
                 output_size: Union[int, Tuple[int, int]]
                 ):
        r"""

        Adaptive channel-wise average-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AdaptiveAvgPool2d`, wrapping it in
        the :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            output_size: the target output size of the image of the form H x W

        """
        super().__init__(
                in_type, 2,
                pool=torch.nn.AdaptiveAvgPool2d(
                    output_size=output_size,
                ),
        )


class PointwiseAdaptiveAvgPool3D(_PointwisePoolND):

    def __init__(self,
                 in_type: FieldType,
                 output_size: Union[int, Tuple[int, int, int]]
                 ):
        r"""

        Adaptive channel-wise average-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AdaptiveAvgPool3d`, wrapping it in
        the :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            output_size: the target output size of the volume of the form H x W x D

        """
        super().__init__(
                in_type, 3,
                pool=torch.nn.AdaptiveAvgPool3d(
                    output_size=output_size,
                ),
        )

class PointwiseAvgPoolAntialiased2D(_PointwiseAvgPoolAntialiasedND):
    
    def __init__(self,
                 in_type: FieldType,
                 sigma: float,
                 stride: Union[int, Tuple[int, int]],
                 padding: Optional[Union[int, Tuple[int, int]]] = None,
                 edge_correction: bool = False,
                 ):
        r"""

        Antialiased channel-wise average-pooling: each channel is treated independently.
        It performs strided convolution with a Gaussian blur filter.
        
        The size of the filter is computed as 3 standard deviations of the Gaussian curve.
        By default, padding is added such that input size is preserved if stride is 1.
        
        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            sigma (float): standard deviation for the Gaussian blur filter
            stride: the stride of the window.
            padding: additional zero padding to be added on both sides

        """
        super().__init__(
            in_type, 2,
            sigma=sigma,
            stride=stride,
            padding=padding,
            edge_correction=edge_correction,
        )

class PointwiseAvgPoolAntialiased3D(_PointwiseAvgPoolAntialiasedND):
    
    def __init__(self,
                 in_type: FieldType,
                 sigma: float,
                 stride: Union[int, Tuple[int, int, int]],
                 padding: Optional[Union[int, Tuple[int, int, int]]] = None,
                 edge_correction: bool = False,
                 ):
        r"""

        Antialiased channel-wise average-pooling: each channel is treated independently.
        It performs strided convolution with a Gaussian blur filter.
        
        The size of the filter is computed as 3 standard deviations of the Gaussian curve.
        By default, padding is added such that input size is preserved if stride is 1.
        
        Inspired by `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            sigma (float): standard deviation for the Gaussian blur filter
            stride: the stride of the window.
            padding: additional zero padding to be added on both sides

        """
        super().__init__(
            in_type, 3,
            sigma=sigma,
            stride=stride,
            padding=padding,
            edge_correction=edge_correction,
        )
        

# for backward compatibility
PointwiseAvgPool = PointwiseAvgPool2D
PointwiseAdaptiveAvgPool = PointwiseAdaptiveAvgPool2D
PointwiseAvgPoolAntialiased = PointwiseAvgPoolAntialiased2D



from typing import Any, List, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from escnn.gspaces import GSpace
from escnn.nn import FieldType, GeometricTensor

from ..equivariant_module import EquivariantModule

__all__ = ["LeakyReLU"]


class LeakyReLU(EquivariantModule):
    def __init__(
        self, in_type: FieldType, negative_slope: float = 0.01, inplace: bool = False
    ):
        r"""

        Module that implements a pointwise LeakyReLU to every channel independently.
        The input representation is preserved by this operation and, therefore, it equals the output
        representation.

        Only representations supporting pointwise non-linearities are accepted as input field type.

        Args:
            in_type (FieldType):  the input field type
            negative_slope (float, optional): Controls the angle of the negative slope (which is used for negative input values). Default: 0.01
            inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

        """

        assert isinstance(in_type.gspace, GSpace)

        super(LeakyReLU, self).__init__()

        for r in in_type.representations:
            assert (
                "pointwise" in r.supported_nonlinearities
            ), 'Error! Representation "{}" does not support "pointwise" non-linearity'.format(
                r.name
            )

        self.space = in_type.gspace
        self.in_type = in_type

        # the representation in input is preserved
        self.out_type = in_type

        self._negative_slope = negative_slope
        self._inplace = inplace

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Applies leaky-ReLU function on the input fields

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map after relu has been applied

        """

        assert (
            input.type == self.in_type
        ), "Error! the type of the input does not match the input type of this module"
        return GeometricTensor(
            F.leaky_relu(
                input.tensor, negative_slope=self._negative_slope, inplace=self._inplace
            ),
            self.out_type,
            input.coords,
        )

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(
        self, atol: float = 1e-6, rtol: float = 1e-5
    ) -> List[Tuple[Any, float]]:
        c = self.in_type.size

        x = torch.randn(3, c, 10, 10)

        x = GeometricTensor(x, self.in_type)

        errors = []

        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))

            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())

            assert torch.allclose(
                out1.tensor, out2.tensor, atol=atol, rtol=rtol
            ), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(
                el, errs.max(), errs.mean(), errs.var()
            )

            errors.append((el, errs.mean()))

        return errors

    def extra_repr(self):
        return "inplace={}, type={}".format(self._inplace, self.in_type)

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.LeakyReLU` module and set to "eval" mode.

        """

        self.eval()

        return torch.nn.LeakyReLU(
            negative_slope=self._negative_slope, inplace=self._inplace
        )



from escnn.nn import FieldType
from escnn.gspaces import GSpace

from numbers import Number

import math

from typing import Tuple, Any

def calc_pool_output_shape(
        d: int,
        pool: Any,
        input_shape: Tuple[int, ...],
        out_type: FieldType,
) -> Tuple[int, ...]:

    assert len(input_shape) == 2 + d, (input_shape, d)

    if hasattr(pool, 'output_size'):
        output_size = get_nd_tuple(pool.output_size, d)
        return (input_shape[0], out_type.size, *output_size)

    b, c, *xyz_i = input_shape

    kernel_size = get_nd_tuple(pool.kernel_size, d)
    padding = get_nd_tuple(pool.padding, d)
    stride = get_nd_tuple(pool.stride, d)
    dilation = get_nd_tuple(getattr(pool, 'dilation', 1), d)

    # See online docs for `torch.nn.MaxPool2D`.
    xyz_o = tuple(
        (xyz_i[i] + 2 * padding[i] - dilation[i] * (kernel_size[i] - 1) - 1) / stride[i] + 1
        for i in range(d)
    )

    ceil_mode = getattr(pool, 'ceil_mode', False)
    round = math.ceil if ceil_mode else math.floor

    xyz_o = tuple(
        int(round(xyz_o[i]))
        for i in range(d)
    )

    return (b, out_type.size, *xyz_o)

def check_dimensions(in_type: FieldType, d: int) -> None:
    assert d in [2, 3], f"Only dimensionality 2 or 3 are currently suported by 'd={d}' found"

    assert isinstance(in_type.gspace, GSpace)
    assert in_type.gspace.dimensionality == d, (in_type.gspace.dimensionality, d)

def check_pointwise_ok(in_type: FieldType) -> None:
    for r in in_type.representations:
        assert 'pointwise' in r.supported_nonlinearities, \
            f"""Error! Representation "{r.name}" does not support pointwise non-linearities
            so it is not possible to pool each channel independently"""

def get_nd_tuple(scalar_or_tuple, d):
    if isinstance(scalar_or_tuple, Number):
        return (scalar_or_tuple,) * d
    else:
        return scalar_or_tuple




# first line: 482
@cache.cache
def thomson_cube_sphere(N: int) -> np.ndarray:
    attempts = 20
    verbose = 0

    # to ensure a determinist behaviour of this method
    rng = np.random.RandomState(42)

    best_U = (24*N) ** 2
    best_X = None
    for i in range(attempts):
        X = _thomson_cube_sphere(N, lr=1e-1, optimizer='Adam', rng=rng, verbose=verbose - 1)
        U = _potential_sphere(X).item()

        if U < best_U:
            best_U = U
            best_X = X

    if verbose > 0:
        print(f'Best Potential: {best_U}')

    return best_X.numpy()



import numpy as np
import torch

from escnn.group import *

from typing import List, Union, Tuple, Callable, Dict, Iterable

from escnn.kernels import *
from escnn.nn import *


__all__ = [
    'NormalizedR3Conv',
]


class GaussianRadialProfileNormalized(KernelBasis):
    
    def __init__(self, radii: List[float], sigma: Union[List[float], float]):
        r"""

        Basis for kernels defined over a radius in :math:`\R^+_0`.

        Each basis element is defined as a Gaussian function.
        Different basis elements are centered at different radii (``rings``) and can possibly be associated with
        different widths (``sigma``).

        More precisely, the following basis is implemented:

        .. math::

            \mathcal{B} = \left\{ b_i (r) :=  \exp \left( \frac{ \left( r - r_i \right)^2}{2 \sigma_i^2} \right) \right\}_i

        In order to build a complete basis of kernels, you should combine this basis with a basis which defines the
        angular profile, see for example :class:`~escnn.kernels.SphericalShellsBasis` or
        :class:`~escnn.kernels.CircularShellsBasis`.

        Args:
            radii (list): centers of each basis element. They should be different and spread to cover all
                domain of interest
            sigma (list or float): widths of each element. Can potentially be different.


        """
        
        if isinstance(sigma, float):
            sigma = [sigma] * len(radii)
        
        assert len(radii) == len(sigma)
        assert isinstance(radii, list)
        
        for r in radii:
            assert r >= 0.
        
        for s in sigma:
            assert s > 0.
        
        super(GaussianRadialProfileNormalized, self).__init__(len(radii), (1, 1))
        
        self.register_buffer('radii', torch.tensor(radii, dtype=torch.float32).reshape(1, -1, 1, 1))
        self.register_buffer('sigma', torch.tensor(sigma, dtype=torch.float32).reshape(1, -1, 1, 1))
    
    def sample(self, radii: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on the discrete set of radii in ``radii``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``radii`` must be an array of shape `(N, 1)`, where `N` is the number of points.

        Args:
            radii (~torch.Tensor): radii where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        assert len(radii.shape) == 2
        assert radii.shape[1] == 1
        S = radii.shape[0]
        
        if out is None:
            out = torch.empty((S, self.dim, self.shape[0], self.shape[1]), device=radii.device, dtype=radii.dtype)
        
        assert out.shape == (S, self.dim, self.shape[0], self.shape[1])
        
        radii = radii.reshape(-1, 1, 1, 1)

        assert not torch.isnan(radii).any()

        d = (self.radii - radii) ** 2

        if radii.requires_grad:
            unnormalized_out = torch.exp(-0.5 * d / self.sigma ** 2, out=out)
            out[:] = unnormalized_out / unnormalized_out.sum(1, keepdims=True)
        else:
            out = torch.exp(-0.5 * d / self.sigma ** 2, out=out)
            out /= out.sum(1, keepdims=True)

        return out
    
    def __getitem__(self, r):
        assert r < self.dim
        return {"radius": self.radii[0, r, 0, 0].item(), "sigma": self.sigma[0, r, 0, 0].item(), "idx": r}
    
    def __eq__(self, other):
        if isinstance(other, GaussianRadialProfile):
            return (
                        torch.allclose(self.radii, other.radii.to(self.radii.device))
                    and torch.allclose(self.sigma, other.sigma.to(self.sigma.device))
            )
        else:
            return False
    
    def __hash__(self):
        return hash(self.radii.cpu().numpy().tobytes()) + hash(self.sigma.cpu().numpy().tobytes())


def normalized_kernels_SO3_act_R3(in_repr: Representation, out_repr: Representation,
                       radii: List[float],
                       # sigma: Union[List[float], float],
                       maximum_frequency: int = None,
                       adjoint: np.ndarray = None,
                       filter: Callable[[Dict], bool] = None
                       ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to continuous rotations, modeled by the
    group :math:`SO(3)`.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s of :class:`~escnn.group.SO3`.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        # sigma (list or float): widths of the rings defining the basis for the radial profile
        adjoint (~numpy.ndarray, optional): 3x3 orthogonal matrix defining a change of basis on the base space

    """
    assert in_repr.group == out_repr.group

    group = in_repr.group

    assert isinstance(group, SO3)

    sigma = 1. / np.sqrt(2. * np.log(2)) / len(radii)

    radial_profile = GaussianRadialProfileNormalized(radii, sigma)

    if maximum_frequency is None:
        max_in_freq = max(freq for freq, in in_repr.irreps)
        max_out_freq = max(freq for freq, in out_repr.irreps)
        maximum_frequency = max_in_freq + max_out_freq

    basis = SteerableKernelBasis(
        SphericalShellsBasis(maximum_frequency, radial_profile, filter=filter),
        in_repr, out_repr,
        RestrictedWignerEckartBasis,
        sg_id='so3'
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (3, 3)
        basis = AdjointBasis(basis, adjoint)

    return basis


_fields_intertwiners_basis_memory_normaliedSO3 = dict()


class NormalizedR3Conv(R3Conv):

    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:

        # sigma = self._sigma
        rings = self._rings
        maximum_frequency=self._maximum_frequency
        # build the key
        key = dict()
        key["rings"] = tuple(rings)
        key["maximum_requency"] = maximum_frequency
        key = tuple(sorted(key.items()))

        if (in_repr.name, out_repr.name) not in _fields_intertwiners_basis_memory_normaliedSO3[key]:
            basis = normalized_kernels_SO3_act_R3(in_repr, out_repr, rings, maximum_frequency=maximum_frequency)
            # store the basis in the dictionary
            _fields_intertwiners_basis_memory_normaliedSO3[key][(in_repr.name, out_repr.name)] = basis
        # return the dictionary with all the basis built for this filter size
        return _fields_intertwiners_basis_memory_normaliedSO3[key][(in_repr.name, out_repr.name)]




# first line: 233
@cache.cache
def _restrict_irrep(irrep_id: Tuple, id, group_class: str, **group_keys) -> Tuple[np.matrix, List[Tuple[Tuple, int]]]:
    
    group = escnn.group.groups_dict[group_class]._generator(**group_keys)
    
    irrep = group.irrep(*irrep_id)

    id = irrep.group._decode_subgroup_id_pickleable(id)

    subgroup, parent, child = group.subgroup(id)
    
    if subgroup.order() == 1:
        # if the subgroup is the trivial group, just return the identity cob and the list of trivial reprs
        return np.eye(irrep.size), [subgroup.trivial_representation.id]*irrep.size
    
    if subgroup.order() > 1:
        # if it is a finite group, we can sample all the element and use the precise method based on Character theory

        representation = {
            g: irrep(parent(g)) for g in subgroup.elements
        }

        # to solve the Sylvester equation and find the change of basis matrix, it is sufficient to sample
        # the generators of the subgroup
        change_of_basis, multiplicities = decompose_representation_finitegroup(
            representation,
            subgroup,
        )

    else:
        # if the group is not finite, we rely on the numerical method which is based on some samples of the group

        representation = lambda g: irrep(parent(g))

        change_of_basis, multiplicities = decompose_representation_general(
            representation,
            subgroup,
        )

    irreps = []
    
    for irr, m in multiplicities:
        irreps += [irr]*m
    
    return change_of_basis, irreps




from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule
from .gaussian_blur import GaussianBlurND, kernel_size_from_radius
from .utils import (
        calc_pool_output_shape, check_dimensions, check_pointwise_ok,
)

import torch
from collections import OrderedDict

from typing import List, Tuple, Any, Union, Optional

_MAX_POOLS = {
        2: torch.nn.MaxPool2d,
        3: torch.nn.MaxPool3d,
}

class _PointwisePoolND(EquivariantModule):

    def __init__(
            self,
            in_type: FieldType,
            d: int,
            pool: torch.nn.Module,
    ):
        r"""

        Channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.MaxPool2D` or :class:`torch.nn.MaxPool3D`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            d (int): dimensionality of the base space (2 for images, 3 for volumes)
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when True, will use ceil instead of floor to compute the output shape

        """

        super().__init__()

        check_dimensions(in_type, d)
        check_pointwise_ok(in_type)

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        self.pool = pool

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map

        """

        assert input.type == self.in_type

        output = self.pool(input.tensor)

        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        return calc_pool_output_shape(self.d, self.pool, input_shape, self.out_type)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # this kind of pooling is not really equivariant so we can not test 
        # equivariance
        pass

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.MaxPool2d` module and set to "eval" mode.

        """

        return self.pool.eval()

class _PointwiseAvgPoolAntialiasedND(EquivariantModule):

    def __init__(
            self,
            in_type: FieldType,
            d: int,
            *,
            sigma: float,
            stride: Union[int, Tuple[int, int]],
            padding: Optional[Union[int, Tuple[int, int]]],
            edge_correction: bool = False,
    ):
        r"""

        Antialiased channel-wise average-pooling: each channel is treated independently.
        It performs strided convolution with a Gaussian blur filter.
        
        The size of the filter is computed as 3 standard deviations of the Gaussian curve.
        By default, padding is added such that input size is preserved if stride is 1.
        
        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            sigma (float): standard deviation for the Gaussian blur filter
            stride: the stride of the window.
            padding: additional zero padding to be added on both sides

        """
        
        super().__init__()

        check_dimensions(in_type, d)
        # Don't need to check that the representation is compatible with 
        # pointwise nonlinearities, see #65.

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self.blur = GaussianBlurND(
                sigma=sigma,
                kernel_size=kernel_size_from_radius(sigma * 3),
                stride=stride,
                padding=padding,
                d=d,
                edge_correction=edge_correction,
                channels=in_type.size,
        )
    
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map

        """
        
        assert input.type == self.in_type

        output = self.blur(input.tensor)

        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        return calc_pool_output_shape(self.d, self.blur, input_shape, self.out_type)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # this kind of pooling is not really equivariant so we can't test 
        # equivariance
        pass

    def export(self):
        return self.blur.eval()

class _PointwiseMaxPoolAntialiasedND(EquivariantModule):

    def __init__(
            self,
            in_type: FieldType,
            d: int,
            *,
            kernel_size: Union[int, Tuple[int, ...]],
            stride: Optional[Union[int, Tuple[int, ...]]],
            padding: Union[int, Tuple[int, ...]],
            ceil_mode: bool,
            sigma: float,
    ):
        r"""

        Anti-aliased version of channel-wise max-pooling (each channel is treated independently).

        The max over a neighborhood is performed pointwise withot downsampling.
        Then, convolution with a gaussian blurring filter is performed before downsampling the feature map.

        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.


        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            d (int): dimensionality of the base space (2 for images, 3 for volumes)
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            sigma (float): standard deviation for the Gaussian blur filter

        """

        super().__init__()

        if ceil_mode:
            from warnings import warn
            warn("The `ceil_mode` argument doesn't do anything, because the stride for the pooling step is always 1.")

        check_dimensions(in_type, d)
        check_pointwise_ok(in_type)

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        pool = _MAX_POOLS[d](
                kernel_size=kernel_size,
                stride=1,
                padding=padding,
        )
        blur = GaussianBlurND(
                sigma=sigma,
                kernel_size=kernel_size_from_radius(sigma * 4),
                stride=stride if stride is not None else kernel_size,
                rel_padding=padding,
                channels=in_type.size,
                d=d,
        )
        self.layers = torch.nn.Sequential(
                OrderedDict([('pool', pool), ('blur', blur)]),
        )

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map

        """

        assert input.type == self.in_type
        
        output = self.layers(input.tensor)

        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        for layer in self.layers:
            input_shape = calc_pool_output_shape(self.d, layer, input_shape, self.out_type)

        return input_shape

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # this kind of pooling is not really equivariant so we can not test 
        # equivariance
        pass

    def export(self):
        return self.layers.eval()




from collections import defaultdict

from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
from torch.nn import Parameter
from typing import List, Tuple, Any, Union
import numpy as np

__all__ = [
    "FieldNorm",
]


class FieldNorm(EquivariantModule):

    def __init__(self,
                 in_type: FieldType,
                 eps: float = 1e-05,
                 affine: bool = True,
                 ):
        r"""

        Normalization module which normalizes each field individually.
        The statistics are only computed over the channels within a single field (not over the batch dimension or
        the spatial dimensions).
        Moreover, this layer does not track running statistics and uses only the current input, so it behaves similarly
        at train and eval time.

        For each individual field, the mean is given by the projection on the subspaces transforming under the trivial
        representation while the variance is the squared norm of the field, after the mean has been subtracted.

        If ``affine = True``, a single scale is learnt per input field and the bias is applied only to the
        trivial irreps (this scale and bias are shared over the spatial dimensions in order to preserve equivariance).

        .. warning::
            If a field is only containing trivial irreps, this layer will just set its values to zero and, possibly,
            replace them with a learnable bias if ``affine = True``.


        Args:
            in_type (FieldType): the input field type
            eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
            affine (bool, optional): if ``True``, this module has learnable affine parameters. Default: ``True``

        """

        assert isinstance(in_type.gspace, GSpace)

        super(FieldNorm, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        self.affine = affine

        self._nfields = None

        # group fields by their type and
        #   - check if fields of the same type are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each type
        self._nfields = defaultdict(int)

        # indices of the channels corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])

        # whether each group of fields is contiguous or not
        self._contiguous = {}

        ntrivials = 0
        position = 0
        last_field = None
        for i, r in enumerate(self.in_type.representations):

            for irr in r.irreps:
                if self.in_type.fibergroup.irrep(*irr).is_trivial():
                    ntrivials += 1

            if r.name != last_field:
                if not r.name in self._contiguous:
                    self._contiguous[r.name] = True
                else:
                    self._contiguous[r.name] = False

            last_field = r.name
            _indices[r.name] += list(range(position, position + r.size))
            self._nfields[r.name] += 1
            position += r.size

        for name, contiguous in self._contiguous.items():

            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[name] = [min(_indices[name]), max(_indices[name]) + 1]
                setattr(self, f"{self._escape_name(name)}_indices", _indices[name])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[name] = torch.LongTensor(_indices[name])

                # register the indices tensors as parameters of this module
                self.register_buffer(f"{self._escape_name(name)}_indices", _indices[name])

        # store the size of each field type
        self._sizes = []

        # store for each field type the indices of the trivial irreps in it
        self._trivial_idxs = {}

        # store for each field type the sizes and the indices of all its irreps, grouped by their size
        self._irreps_sizes = {}

        self._has_trivial = {}

        # for each different representation in the input type
        for r in self.in_type._unique_representations:
            p = 0
            trivials = []

            # mask containing the location of the trivial irreps in the irrep decomposition of the representation
            S = np.zeros((r.size, r.size))

            # find all trivial irreps occurring in the representation
            for i, irr in enumerate(r.irreps):
                irr = self.in_type.fibergroup.irrep(*irr)
                if irr.is_trivial():
                    trivials.append(p)
                    S[p, p] = 1.

                p += irr.size

            name = r.name
            self._sizes.append((name, r.size))

            self._has_trivial[name] = len(trivials) > 0

            if self._has_trivial[name]:
                # averaging matrix which computes the expectation of a input vector, i.e. projects it in the trivial
                # subspace by masking out all non-trivial irreps
                P = r.change_of_basis @ S @ r.change_of_basis_inv
                self.register_buffer(f'{self._escape_name(name)}_avg', torch.tensor(P, dtype=torch.float))

                Q = torch.tensor(r.change_of_basis, dtype=torch.float)[:, trivials]
                self.register_buffer(f'{self._escape_name(name)}_change_of_basis', Q)

            if self.affine:
                # scale all dimensions of the same field by the same weight
                weight = Parameter(torch.ones((self._nfields[r.name], 1)), requires_grad=True)
                self.register_parameter(f'{self._escape_name(name)}_weight', weight)
                if self._has_trivial[name]:
                    # the bias is applied only to the trivial channels
                    bias = Parameter(torch.zeros((self._nfields[r.name], len(trivials))), requires_grad=True)
                    self.register_parameter(f'{self._escape_name(name)}_bias', bias)

        self.eps = eps

    def reset_parameters(self):
        if self.affine:
            for name, size in self._sizes:
                weight = getattr(self, f"{self._escape_name(name)}_weight")
                weight.data.fill_(1)
                if hasattr(self, f"{self._escape_name(name)}_bias"):
                    bias = getattr(self, f"{self._escape_name(name)}_bias")
                    bias.data.fill_(0)

    def reset_running_stats(self):
        pass

    def _estimate_stats(self, slice, name: str):

        if self._has_trivial[name]:
            P = getattr(self, f'{self._escape_name(name)}_avg')

            # compute the mean
            means = torch.einsum(
                'ij,bcj...->bci...',
                P,
                slice.detach()
            )
            centered = slice - means
        else:
            means = None
            centered = slice

        # Center the data and compute the variance
        # N.B.: we implicitly assume the dimensions to be iid,
        # i.e. the covariance matrix is a scalar multiple of the identity
        # vars = centered.var(dim=2, unbiased=False, keepdim=True).detach()
        vars = (centered**2).mean(dim=2, keepdim=True).detach()

        return means, vars

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        Normalize the input feature map

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map

        """

        assert input.type == self.in_type

        coords = input.coords
        input = input.tensor
        b, c = input.shape[:2]
        spatial_dims = input.shape[2:]

        output = torch.empty_like(input)

        # iterate through all field types
        for name, size in self._sizes:
            indices = getattr(self, f"{self._escape_name(name)}_indices")

            if self._contiguous[name]:
                slice = input[:, indices[0]:indices[1], ...]
            else:
                slice = input[:, indices, ...]

            slice = slice.view(b, -1, size, *spatial_dims)

            means, vars = self._estimate_stats(slice, name)

            if self._has_trivial[name]:
                # center data by subtracting the mean
                slice = slice - means

            # normalize dividing by the std and multiply by the new scale
            if self.affine:
                weight = getattr(self, f"{self._escape_name(name)}_weight").view(1, self._nfields[name], 1, *(1,)*len(spatial_dims))
            else:
                weight = 1.

            # compute the scalar multipliers needed
            scales = weight / (vars + self.eps).sqrt()
            # scales[vars < self.eps] = 0

            # print(name, size, indices, self._has_trivial[name])
            # print(slice.shape, scales.shape)
            # if not self.training:
            #     np.set_printoptions(precision=5, suppress=True, threshold=1000000, linewidth=3000)
            #     print(scales.detach().cpu().numpy().reshape(scales.shape[0], -1).T)

            # scale features
            slice = slice * scales

            # shift the features with the learnable bias
            if self.affine and self._has_trivial[name]:
                bias = getattr(self, f"{self._escape_name(name)}_bias")
                Q = getattr(self, f'{self._escape_name(name)}_change_of_basis')
                slice = slice + torch.einsum(
                    'ij,cj->ci',
                    Q,
                    bias
                ).view(1, bias.shape[0], Q.shape[0], *(1,) * len(spatial_dims))

            # needed for PyTorch's adaptive mixed precision
            slice = slice.to(output.dtype)

            if not self._contiguous[name]:
                output[:, indices, ...] = slice.view(b, -1, *spatial_dims)
            else:
                output[:, indices[0]:indices[1], ...] = slice.view(b, -1, *spatial_dims)

        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        assert len(input_shape) > 1, input_shape
        assert input_shape[1] == self.in_type.size, input_shape

        return (input_shape[0], self.out_type.size, *input_shape[2:])

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(NormBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass

    def _escape_name(self, name: str):
        return name.replace('.', '^')

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')

        main_str = self._get_name() + '('
        if len(extra_lines) == 1:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(extra_lines) + '\n'

        main_str += ')'
        return main_str

    def extra_repr(self):
        return '{in_type}, eps={eps}, affine={affine}' \
            .format(**self.__dict__)

    def export(self):
        raise NotImplementedError()





from .field_type import FieldType, FourierFieldType
from .geometric_tensor import GeometricTensor, tensor_directsum
from .grid_tensor import GridTensor

from .modules import *
from .modules import __all__ as modules_list

__all__ = [
    "FieldType",
    "FourierFieldType",
    "GeometricTensor",
    "GridTensor",
    "tensor_directsum",
    *modules_list,
    "init",
]



# first line: 633
@cache.cache(ignore=['octa'])
def _build_octa_irrep(octa: Octahedral, l: int):
    
    if l == -1:
        
        # matrix coefficients from https://arxiv.org/pdf/1110.6376.pdf
        
        # the matrix coefficients there are expressed wrt a different set of generators
        # we fist build this set of generators
        
        r3 = octa.generators[0]
        r = r3 @ r3 @ r3
        
        k = octa.elements[0]
        s = octa.generators[1]
        t = ~s @ k @ s @ r
        
        # Representation of `t`
        rho_t = np.array([
            [1., 0., 0.],
            [0., 0., 1.],
            [0., 1., 0.],
        ])
        
        # Representation of `k`
        rho_k = np.array([
            [1., 0., 0.],
            [0., -1., 0.],
            [0., 0., -1.],
        ])
        
        # Representation of `s`
        rho_s = np.array([
            [0., 1., 0.],
            [0., 0., 1.],
            [1., 0., 0.],
        ])
        
        #  https://arxiv.org/pdf/1110.6376.pdf defines the irrep `l = 1` (denoted by 3 there) as our
        #  `standard_representation`, which is expressed on a different basis than the Wigner D matrix with l=1.
        # Since `l=-1` (their 3') is defined as the tensor product between `l=1` and `l=3` (their 1')
        # we apply the inverse change of basis used in `standard_representation` to ensure that
        # `-1 = 1 \tensor 3` for us as well
        
        change_of_basis = np.array([
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0]
        ])
        
        rho_t = change_of_basis.T @ rho_t @ change_of_basis
        rho_k = change_of_basis.T @ rho_k @ change_of_basis
        rho_s = change_of_basis.T @ rho_s @ change_of_basis
        
        generators = [
            (t, rho_t),
            (k, rho_k),
            (s, rho_s),
        ]
        
        return generate_irrep_matrices_from_generators(octa, generators)

    elif l == 2:

        # matrix coefficients from https://arxiv.org/pdf/1110.6376.pdf

        # the matrix coefficients there are expressed wrt a different set of generators
        # we fist build this set of generators

        r3 = octa.generators[0]
        r = r3 @ r3 @ r3

        k = octa.elements[0]
        s = octa.generators[1]
        t = ~s @ k @ s @ r

        # Representation of `t`
        rho_t = np.array([
            [0., 1.],
            [1., 0.],
        ])

        # Representation of `k`
        rho_k = np.array([
            [1., 0.],
            [0., 1.],
        ])

        # Representation of `s`
        rho_s = 0.5 * np.array([
            [-1., -np.sqrt(3)],
            [np.sqrt(3), -1.],
        ])

        generators = [
            (t, rho_t),
            (k, rho_k),
            (s, rho_s),
        ]
        
        return generate_irrep_matrices_from_generators(octa, generators)

    elif l == 3:

        # matrix coefficients from https://arxiv.org/pdf/1110.6376.pdf

        # the matrix coefficients there are expressed wrt a different set of generators
        # we fist build this set of generators

        r3 = octa.generators[0]
        r = r3 @ r3 @ r3

        k = octa.elements[0]
        s = octa.generators[1]
        t = ~s @ k @ s @ r

        # Representation of `t`
        rho_t = np.array([[-1.]])

        # Representation of `k`
        rho_k = np.array([[1.]])

        # Representation of `s`
        rho_s = np.array([[1.]])

        generators = [
            (t, rho_t),
            (k, rho_k),
            (s, rho_s),
        ]

        return generate_irrep_matrices_from_generators(octa, generators)

    else:
        raise ValueError()



# first line: 671
@cache.cache(ignore=['ico'])
def _build_ico_irrep(ico: Icosahedral, l: int):
    
    if l == 3:
        
        # Representation of the generator of the cyclic subgroup of order 5
        rho_p = np.zeros((3, 3))
        
        rho_p[0, 0] = rho_p[1, 1] = np.cos(144 / 180. * np.pi)
        rho_p[1, 0] = np.sin(144 / 180. * np.pi)
        rho_p[0, 1] = -np.sin(144 / 180. * np.pi)
        rho_p[2, 2] = 1.
        
        # Representation of the generator of the cyclic subgroup of order 2
        rho_q = np.zeros((3, 3))
        rho_q[0, 0] = 1. / np.sqrt(5)
        rho_q[0, 2] = - 2. / np.sqrt(5)
        rho_q[1, 1] = - 1
        rho_q[2, 0] = - 2. / np.sqrt(5)
        rho_q[2, 2] = - 1. / np.sqrt(5)
        
        generators = [
            (ico._generators[0], rho_p),
            (ico._generators[1], rho_q),
        ]
        
        return generate_irrep_matrices_from_generators(ico, generators)

    elif l == 4:

        # Representation of the generator of the cyclic subgroup of order 5
        rho_p = np.zeros((4, 4))

        rho_p[0, 0] = rho_p[1, 1] = np.cos(72 / 180. * np.pi)
        rho_p[1, 0] = np.sin(72 / 180. * np.pi)
        rho_p[0, 1] = -np.sin(72 / 180. * np.pi)

        rho_p[2, 2] = rho_p[3, 3] = np.cos(144 / 180. * np.pi)
        rho_p[3, 2] = np.sin(144 / 180. * np.pi)
        rho_p[2, 3] = -np.sin(144 / 180. * np.pi)

        # Representation of the generator of the cyclic subgroup of order 2
        rho_q = np.zeros((4, 4))
        rho_q[0, 2] = -1
        rho_q[1, 1] = 2. / np.sqrt(5)
        rho_q[1, 3] = 1. / np.sqrt(5)
        rho_q[2, 0] = -1
        rho_q[3, 1] = 1. / np.sqrt(5)
        rho_q[3, 3] = -2. / np.sqrt(5)

        generators = [
            (ico._generators[0], rho_p),
            (ico._generators[1], rho_q),
        ]

        return generate_irrep_matrices_from_generators(ico, generators)
    else:
        raise ValueError()



# first line: 172
@cache.cache
def _find_tensor_decomposition(J: Tuple, l: Tuple, group_class: str, **group_keys) -> List[Tuple[Tuple, int]]:
    G = escnn.group.groups_dict[group_class]._generator(**group_keys)
    
    psi_J = G.irrep(*J)
    psi_l = G.irrep(*l)
    
    irreps = []
    
    size = 0
    for psi_j in G.irreps():
        CG = G._clebsh_gordan_coeff(psi_J, psi_l, psi_j)
        
        S = CG.shape[-2]
        
        if S > 0:
            irreps.append((psi_j.id, S))
        
        size += psi_j.size * S

    # check that size == psi_J.size * psi_l.size
    
    if size < psi_J.size * psi_l.size:
        from textwrap import dedent
        message = dedent(f"""
            Error! Did not find sufficient irreps to complete the decomposition of the tensor product of '{psi_J.name}' and '{psi_l.name}'.
            It is likely this happened because not sufficiently many irreps in '{G}' have been instantiated.
            Try instantiating more irreps and then repeat this call.
            The sum of the sizes of the irreps found is {size}, but the representation has size {psi_J.size * psi_l.size}.
        """)
        raise InsufficientIrrepsException(G, message)

    assert size <= psi_J.size * psi_l.size, f"""
        Error! Found too many irreps in the the decomposition of the tensor product of '{psi_J.name}' and '{psi_l.name}'.
        This should never happen!
    """

    return irreps



from .norm_max import (
        NormMaxPool2D, NormMaxPool,
        NormMaxPool3D,
)

from .pointwise_max import (
        PointwiseMaxPool2D, PointwiseMaxPool,
        PointwiseMaxPool3D,

        PointwiseAdaptiveMaxPool2D, PointwiseAdaptiveMaxPool,
        PointwiseAdaptiveMaxPool3D,

        PointwiseMaxPoolAntialiased2D, PointwiseMaxPoolAntialiased,
        PointwiseMaxPoolAntialiased3D,
)
from .pointwise_avg import (
        PointwiseAvgPool2D, PointwiseAvgPool,
        PointwiseAvgPool3D,

        PointwiseAdaptiveAvgPool2D, PointwiseAdaptiveAvgPool,
        PointwiseAdaptiveAvgPool3D,
        
        PointwiseAvgPoolAntialiased2D, PointwiseAvgPoolAntialiased,
        PointwiseAvgPoolAntialiased3D,
)

__all__ = [
    "NormMaxPool2D", "NormMaxPool",
    "NormMaxPool3D",
    "PointwiseMaxPool2D", "PointwiseMaxPool",
    "PointwiseMaxPool3D",
    "PointwiseMaxPoolAntialiased2D", "PointwiseMaxPoolAntialiased",
    "PointwiseMaxPoolAntialiased3D",
    "PointwiseAvgPool2D", "PointwiseAvgPool",
    "PointwiseAvgPoolAntialiased2D", "PointwiseAvgPoolAntialiased",
    "PointwiseAdaptiveAvgPool2D", "PointwiseAdaptiveAvgPool",
    "PointwiseAdaptiveAvgPool3D",
    "PointwiseAdaptiveMaxPool2D", "PointwiseAdaptiveMaxPool",
    "PointwiseAdaptiveMaxPool3D",
    ###### 3D ##################
    "PointwiseAvgPool3D",
    "PointwiseAvgPoolAntialiased3D",
]






from escnn.group import *
from escnn.gspaces import *
from escnn.nn import FourierFieldType
from escnn.nn import GeometricTensor

from .fourier import FourierPointwise

from typing import List, Tuple, Any

__all__ = ["QuotientFourierPointwise", "QuotientFourierELU"]


class QuotientFourierPointwise(FourierPointwise):
    
    def __init__(
            self,
            gspace: GSpace,
            subgroup_id: Tuple,
            channels: int,
            irreps: List,
            *grid_args,
            grid: List[GroupElement] = None,
            function: str = 'p_relu',
            inplace: bool = True,
            out_irreps: List = None,
            normalize: bool = True,
            **grid_kwargs
    ):
        r"""

        Perform an Inverse Fourier Transform to sample the input features on a *quotient space* :math:`X`, apply the
        pointwise non-linearity in the spatial domain (Dirac-delta basis) and, finally, compute the Fourier Transform
        to obtain irreps coefficients.

        The quotient space used is isomorphic to :math:`X \cong G / H` where :math:`G` is ```gspace.fibergroup``` while
        :math:`H` is the subgroup of :math:`G` idenitified by ```subgroup_id```; see
        :meth:`~escnn.group.Group.subgroup` and :meth:`~escnn.group.Group.homspace`
        
        .. warning::
            This operation is only *approximately* equivariant and its equivariance depends on the sampling grid and the
            non-linear activation used, as well as the original band-limitation of the input features.
            
        The same function is applied to every channel independently.
        By default, the input representation is preserved by this operation and, therefore, it equals the output
        representation.
        Optionally, the output can have a different band-limit by using the argument ``out_irreps``.

        The class first constructs a band-limited quotient representation of ```gspace.fibergroup``` using
        :meth:`escnn.group.Group.spectral_quotient_representation`.
        The band-limitation of this representation is specified by ```irreps``` which should be a list containing
        a list of ids identifying irreps of ```gspace.fibergroup``` (see :attr:`escnn.group.IrreducibleRepresentation.id`).
        This representation is used to define the input and output field types, each containing ```channels``` copies
        of a feature field transforming according to this representation.
        A feature vector transforming according to such representation is interpreted as a vector of coefficients
        parameterizing a function over the group using a band-limited Fourier basis.

        .. note::
            Instead of building the list ``irreps`` manually, most groups implement a method ``bl_irreps()`` which can be
            used to generate this list with through a simpler interface. Check each group's documentation.

        To approximate the Fourier transform, this module uses a finite number of samples from the group.
        The set of samples to be used can be specified through the parameter ```grid``` or by the ```grid_args``` and
        ```grid_kwargs``` which will then be passed to the method :meth:`~escnn.group.Group.grid`.
        
        .. warning ::
            By definition, an homogeneous space is invariant under a right action of the subgroup :math:`H`.
            That means that a feature representing a function over a homogeneous space :math:`X \cong G/H`, when
            interpreted as a function over :math:`G` (as we do here when sampling), the function will be constant along
            each coset, i.e. :math:`f(gh) = f(g)` if :math:`g \in G, h\in H`.
            An approximately uniform sampling grid over :math:`G` creates an approximately uniform grid over :math:`G/H`
            through projection but might contain redundant elements (if the grid contains :math:`g \in G`, any element
            :math:`gh` in the grid will be redundant).
            It is therefore advised to create a grid directly in the quotient space, e.g. using
            :meth:`escnn.group.SO3.sphere_grid`, :meth:`escnn.group.O3.sphere_grid`.
            We do not support yet a general method and interface to generate grids over any homogeneous space for any
            group, so you should check each group's methods.

        Args:
            gspace (GSpace):  the gspace describing the symmetries of the data. The Fourier transform is
                              performed over the group ```gspace.fibergroup```
            subgroup_id (tuple): identifier of the subgroup :math:`H` to construct the quotient space
            channels (int): number of independent fields in the input `FieldType`
            irreps (list): list of irreps' ids to construct the band-limited representation
            *grid_args: parameters used to construct the discretization grid
            grid (list, optional): list containing the elements of the group to use for sampling. Optional (default ``None``).
            function (str): the identifier of the non-linearity.
                    It is used to specify which function to apply.
                    By default (``'p_relu'``), ReLU is used.
            inplace (bool): applies the non-linear activation in-place. Default: `True`
            out_irreps (list, optional): optionally, one can specify a different band-limiting in output
            normalize (bool, optional): if ``True``, the rows of the IFT matrix (and the columns of the FT matrix) are normalized. Default: ``True``
            **grid_kwargs: keyword parameters used to construct the discretization grid
            
        """
        from warnings import warn
        warn("The `QuotientFourierPointwise` module is deprecated.  Instead, create a \"quotient\" field type by passing the `subgroup_id` argument to `FourierFieldType`, then use `FourierPointwise` on that field.", DeprecationWarning)

        in_type = FourierFieldType(
                gspace, channels, irreps,
                subgroup_id=subgroup_id,
        )
        out_type = FourierFieldType(
                gspace, channels, out_irreps or irreps,
                subgroup_id=subgroup_id,
        )

        if grid is None:
            grid = gspace.fibergroup.grid(*grid_args, **grid_kwargs)

        super().__init__(
                in_type, grid, function,
                out_type=out_type,
                inplace=inplace,
                normalize=normalize,
        )


class QuotientFourierELU(QuotientFourierPointwise):
    
    def __init__(self,
                 gspace: GSpace,
                 subgroup_id: Tuple,
                 channels: int,
                 irreps: List,
                 *grid_args,
                 grid: List[GroupElement] = None,
                 inplace: bool = True,
                 out_irreps: List = None,
                 normalize: bool = True,
                 **grid_kwargs
                 ):
        r"""

        Applies a Inverse Fourier Transform to sample the input features on a quotient space, apply ELU point-wise
        in the spatial domain (Dirac-delta basis) and, finally, computes the Fourier Transform to obtain irreps
        coefficients. See :class:`~escnn.nn.QuotientFourierPointwise` for more details.

        Args:
            gspace (GSpace):  the gspace describing the symmetries of the data. The Fourier transform is
                              performed over the group ```gspace.fibergroup```
            subgroup_id (tuple): identifier of the subgroup :math:`H` to construct the quotient space
            channels (int): number of independent fields in the input `FieldType`
            irreps (list): list of irreps' ids to construct the band-limited representation
            *grid_args: parameters used to construct the discretization grid
            grid (list, optional): list containing the elements of the group to use for sampling. Optional (default ``None``).
            inplace (bool): applies the non-linear activation in-place. Default: ``True``
            out_irreps (list, optional): optionally, one can specify a different band-limiting in output
            normalize (bool, optional): if ``True``, the rows of the IFT matrix (and the columns of the FT matrix) are normalized. Default: ``True``
            **grid_kwargs: keyword parameters used to construct the discretization grid

        """
        
        super(QuotientFourierELU, self).__init__(
            gspace, subgroup_id, channels, irreps, *grid_args,
            function='p_elu',
            inplace=inplace,
            grid=grid,
            out_irreps=out_irreps,
            normalize=normalize,
            **grid_kwargs
        )




import numpy as np
import torch

from escnn.nn import FourierFieldType, GeometricTensor, GridTensor
from escnn.group import Group, GroupElement
from torch.nn import Module

from typing import Sequence, Optional

__all__ = ['FourierTransform', 'InverseFourierTransform']

# Docs:
# - Low level class, meant for building other modules
# - If use directly, possible to break equivariance

class InverseFourierTransform(Module):

    def __init__(
            self,
            in_type: FourierFieldType,
            out_grid: Sequence[GroupElement],
            *,
            normalize: bool = True,
    ):
        super().__init__()

        assert isinstance(in_type, FourierFieldType)

        self.in_type = in_type
        self.out_grid = list(out_grid)

        A = _build_ift(in_type, out_grid, normalize)
        A = torch.tensor(A, dtype=torch.get_default_dtype())

        self.register_buffer('A', A)

    def forward(self, input: GeometricTensor) -> GridTensor:
        assert input.type == self.in_type
        
        x_hat = input.tensor.view(
                input.shape[0],
                self.in_type.channels,
                self.in_type.rho.size,
                *input.shape[2:],
        )

        x = torch.einsum('bcf...,gf->bcg...', x_hat, self.A)

        return GridTensor(x, self.out_grid, input.coords)

class FourierTransform(Module):

    def __init__(
            self,
            in_grid,
            out_type,
            *,
            extra_irreps: Optional[list] = None,
            normalize: bool = True,
    ):
        super().__init__()

        assert isinstance(out_type, FourierFieldType)

        self.in_grid = in_grid
        self.out_type = out_type

        if extra_irreps is None:
            ift_type = out_type
        else:
            extra_irreps = [
                    x
                    for x in extra_irreps
                    if x not in out_type.bl_irreps
            ]
            ift_type = FourierFieldType(
                    out_type.gspace,
                    out_type.channels,
                    out_type.bl_irreps + extra_irreps,
                    subgroup_id=out_type.subgroup_id,
            )

        A = _build_ift(ift_type, in_grid, normalize)

        eps = 1e-8
        n = ift_type.rho.size
        Ainv = np.linalg.inv(A.T @ A + eps * np.eye(n)) @ A.T

        if extra_irreps is not None:
            Ainv = Ainv[:out_type.rho.size, :]

        Ainv = torch.tensor(Ainv, dtype=torch.get_default_dtype())
        
        self.register_buffer('Ainv', Ainv)

    def forward(self, input: GridTensor) -> GeometricTensor:
        assert input.grid == self.in_grid

        y = input.tensor

        y_hat = torch.einsum('bcg...,fg->bcf...', y, self.Ainv)

        y_hat = y_hat.reshape(y.shape[0], self.out_type.size, *y.shape[3:])

        return GeometricTensor(y_hat, self.out_type, input.coords)


def _build_ift(in_type: FourierFieldType, out_grid, normalize: bool):
    """
    Create a matrix that will apply an inverse Fourier transform to a feature 
    vector of the given *in_type*.
    """
    assert isinstance(in_type, FourierFieldType)

    G = in_type.fibergroup

    if in_type.subgroup_id is None:
        kernel = _build_regular_kernel(G, in_type.bl_irreps)
    else:
        kernel = _build_quotient_kernel(G, in_type.subgroup_id, in_type.bl_irreps)

    assert kernel.shape[0] == in_type.rho.size

    if normalize:
        kernel = kernel / np.linalg.norm(kernel)

    kernel = kernel.reshape(-1, 1)
    
    return np.concatenate(
        [
            in_type.rho(g) @ kernel
            for g in out_grid
        ], axis=1
    ).T

def _build_regular_kernel(G: Group, irreps: list[tuple]):
    kernel = []
    
    for irr in irreps:
        irr = G.irrep(*irr)
        
        c = int(irr.size//irr.sum_of_squares_constituents)
        k = irr(G.identity)[:, :c] * np.sqrt(irr.size)
        kernel.append(k.T.reshape(-1))
    
    kernel = np.concatenate(kernel)
    return kernel

def _build_quotient_kernel(G: Group, subgroup_id: tuple, irreps: list[tuple]):
    kernel = []
    
    X: HomSpace = G.homspace(subgroup_id)
    
    for irr in irreps:
        k = X._dirac_kernel_ft(irr, X.H.trivial_representation.id)
        kernel.append(k.T.reshape(-1))
    
    kernel = np.concatenate(kernel)
    return kernel




from escnn.nn import GeometricTensor
from escnn.nn import FieldType
from escnn.gspaces import GSpace
from .equivariant_module import EquivariantModule

from itertools import product, repeat
from typing import Tuple

import torch
import numpy as np

import math

__all__ = ["MaskModule"]


def build_mask(
        s,
        dim: int = 2,
        margin: float = 2.0,
        sigma: float = 2.0,
        dtype=torch.float32,
):
    mask = torch.zeros(1, 1, *repeat(s, dim), dtype=dtype)
    c = (s-1) / 2
    t = (c - margin/100.*c)**2
    for k in product(range(s), repeat=dim):
        r = sum((x - c)**2 for x in k)
        if r > t:
            mask[(..., *k)] = math.exp((t - r) / sigma**2)
        else:
            mask[(..., *k)] = 1.
    return mask


class MaskModule(EquivariantModule):

    def __init__(
            self,
            in_type: FieldType,
            S: int,
            margin: float = 0.,
            sigma: float = 2.,
    ):
        r"""
        
        Performs an element-wise multiplication of the input with a *mask* of shape :math:`S^n`, where :math:`n` is the
        dimensionality of the underlying space.
        
        The mask has value :math:`1` in all pixels with distance smaller than
        :math:`\frac{S - 1}{2} \times (1 - \frac{\mathrm{margin}}{100})` from the center of the mask and :math:`0`
        elsewhere. Values change smoothly between the two regions.
        
        This operation is useful to remove from an input image or feature map all the part of the signal defined on the
        pixels which lay outside the circle inscribed in the grid.
        Because a rotation would move these pixels outside the grid, this information would anyways be
        discarded when rotating an image. However, allowing a model to use this information might break the guaranteed
        equivariance as rotated and non-rotated inputs have different information content.
        
        .. note::
            The input tensors provided to this module must have the following dimensions: :math:`B \times C \times S^n`,
            where :math:`B` is the minibatch dimension, :math:`C` is the channels dimension, and :math:`S^n` are the
            :math:`n` spatial dimensions (corresponding to the Euclidean basespace :math:`\R^n`) associated with the
            given input field type, i.e. ``in_type.gspace.dimensionality``.
            Each Euclidean dimension must be of size :math:`S`.

            For example, if :math:`S=10` and the ``in_type.gspace.dimensionality=2``, then the input tensors
            should be of size :math:`B \times C \times 10 \times 10`.  If ``in_type.gspace.dimensionality=3``
            instead, then the input tensors should be of size
            :math:`B \times C \times 10 \times 10 \times 10`.
        
        Args:
            in_type (FieldType): input field type
            S (int): the shape of the mask and the expected inputs
            margin (float, optional): margin around the mask in percentage with respect to the radius of the mask
            sigma (float, optional): how quickly masked pixels should approach 0.  This can be thought of a standard
                deviation in units of pixels/voxels.  For example, the default value of 2 means that only 5% of the
                original signal will remain 4 px into the masked region.
        
        """
        super(MaskModule, self).__init__()

        self.dim: int = in_type.gspace.dimensionality
        self.S: int = S

        self.margin = margin
        self.mask = torch.nn.Parameter(
                build_mask(S, dim=self.dim, margin=margin, sigma=sigma),
                requires_grad=False,
        )

        self.in_type = self.out_type = in_type
        self.space: GSpace = in_type.gspace

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        assert input.type == self.in_type

        assert input.tensor.shape[2:] == self.mask.shape[2:]

        out = input.tensor * self.mask
        return GeometricTensor(out, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        return input_shape

    def check_equivariance(self, atol: float = 0.1, rtol: float = 0.1, assertion: bool = True, verbose: bool = True, device: str = 'cpu'):

        # np.set_printoptions(precision=5, threshold=30 *self.in_type.size**2, suppress=False, linewidth=30 *self.in_type.size**2)

        feature_map_size = self.S

        c = self.in_type.size

        from tqdm import tqdm
        from skimage.transform import resize

        import scipy
        x = scipy.datasets.face().transpose((2, 0, 1))[np.newaxis, 0:c, :, :]

        x = resize(
            x,
            (x.shape[0], x.shape[1],) + (feature_map_size,)*self.dim,
            anti_aliasing=True
        )

        assert x.shape[0] == 1, x.shape

        x = x / 255.0 - 0.5

        if x.shape[1] < c:
            to_stack = [x for i in range(c // x.shape[1])]
            if c % x.shape[1] > 0:
                to_stack += [x[:, :(c % x.shape[1]), ...]]

            x = np.concatenate(to_stack, axis=1)

        x = torch.FloatTensor(x)
        x = self.in_type(x)

        with torch.no_grad():
            self.to(device)

            gx = self.in_type(torch.cat([x.transform(el).tensor for el in self.space.testing_elements], dim=0))

            gx = gx.to(device)
            outs_2 = self(gx)
            outs_2 = outs_2.tensor.detach().cpu().numpy()
            assert outs_2.shape[-self.dim:] == (feature_map_size, ) * self.dim, (outs_2.shape, feature_map_size)

            out_1 = self(x.to(device)).to('cpu')
            outs_1 = torch.cat([out_1.transform(el).tensor for el in self.space.testing_elements], dim=0)
            del out_1
            outs_1 = outs_1.detach().cpu().numpy()
            assert outs_1.shape[-self.dim:] == (feature_map_size, ) * self.dim, (outs_1.shape, feature_map_size)

            errors = []

            for i, el in tqdm(enumerate(self.space.testing_elements)):

                # out1 = shrink(out_1.transform(el), last_downsampling).tensor.detach().numpy()

                out1 = outs_1[i:i+1]
                out2 = outs_2[i:i+1]

                b, c = out2.shape[:2]
                spatial_dims = out2.shape[2:]

                center_mask = np.stack(np.meshgrid(*[np.arange(0, _w) - _w // 2 for _w in spatial_dims]), axis=0)
                assert center_mask.shape == (len(spatial_dims), *spatial_dims), (center_mask.shape, *spatial_dims)
                center_mask = (center_mask ** 2).sum(0) < (spatial_dims[0] / 4) ** 2

                out1 = out1[..., center_mask]
                out2 = out2[..., center_mask]

                out1 = out1.reshape(-1)
                out2 = out2.reshape(-1)

                errs = np.abs(out1 - out2)

                esum = np.maximum(np.abs(out1), np.abs(out2))
                esum[esum == 0.0] = 1

                relerr = errs / esum

                if verbose:
                    print(el, relerr.max(), relerr.mean(), relerr.var(), errs.max(), errs.mean(), errs.var())

                tol = rtol * esum + atol

                if np.any(errs > tol) and verbose:
                    print(out1[errs > tol])
                    print(out2[errs > tol])
                    print(tol[errs > tol])

                if assertion:
                    assert np.all(
                        errs < tol), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(
                        el, errs.max(), errs.mean(), errs.var())

                errors.append((el, errs.mean()))

        return errors





from .utils import get_nd_tuple

import torch
import torch.nn.functional as F

from torch.nn.modules import Module
from torch.nn.modules.lazy import LazyModuleMixin
from torch.nn.parameter import UninitializedBuffer, is_lazy

from typing import Optional, Union, Tuple

class GaussianBlurND(LazyModuleMixin, Module):

    def __init__(
            self,
            *, 
            d: int,
            sigma: float,
            kernel_size: int,
            stride: Union[int, Tuple[int, ...]] = 1,
            padding: Optional[Union[int, Tuple[int, ...]]] = None,
            rel_padding: Optional[Union[int, Tuple[int, ...]]] = None,
            edge_correction: bool = False,
            channels: Optional[int] = None,
    ):
        """
        Apply a Gaussian blur to the input.

        This is equivalent to a depth-wise convolution with a Gaussian filter.

        Args:
            sigma (float): Standard deviation of the Gaussian making up the 
                blur filter.

            kernel_size (int): Size of the convolutional filter used to apply 
                the blur.  Note that this should be related to the value of 
                *sigma*; larger standard deviations require larger kernels to 
                overlap the same density.  You can think of the Gaussian blur 
                as being truncated to zero beyond the bounds of the filter.

            stride (int): Stride of the convolutional filter used to apply the 
                blur.

            padding: Implicit zero padding to be added on all sides of the 
                input, without regard to the size of the filter.  It is an 
                error to specify *padding* and *rel_padding*.

            rel_padding: Implicit zero padding to be added on all sides of the 
                input, treating the filter as if it were 1x1 (or 1x1x1, etc.), 
                no matter what size it really is.  This means that the shape of 
                the output tensor is independent of the filter size.  This is 
                helpful when, for example, the *kernel_size* argument is 
                dynamically calculated as a function of *sigma*.  It is an 
                error to specify *padding* and *rel_padding*.
            
            d (int): Dimensionality of the base space (2 for images, 3 for 
                volumes).

            channels (int): Channel dimension of the input.  If specified, the 
                convolutional filter can be constructed immediately.  
                Otherwise, it will be constructed during the first forward 
                pass.  Really, the only reason to specify this parameter is to 
                double check that the input has the expected number of 
                channels.
        """
        super().__init__()

        assert sigma > 0.

        if padding is not None and rel_padding is not None:
            raise ValueError("can't specify `padding` and `rel_padding`")

        self.sigma = sigma
        self.kernel_size = kernel_size
        self.stride = stride
        self.edge_correction = edge_correction
        self.d = d
        self.conv = getattr(F, f'conv{d}d')

        if padding is not None:
            self.padding = padding
        else:
            half_kernel_size = (kernel_size - 1) // 2
            self.padding = tuple(
                    p + half_kernel_size
                    for p in get_nd_tuple(rel_padding or 0, d)
            )

        if channels is not None:
            filter_ = make_gaussian_filter(sigma, kernel_size, d, channels)
        else:
            filter_ = UninitializedBuffer()

        self.register_buffer('filter', filter_, persistent=False)

        if edge_correction:
            self.register_buffer('weights', UninitializedBuffer(), persistent=False)

    def __repr__(self):
        return f'{self.__class__.__name__}(sigma={self.sigma}, kernel_size={self.kernel_size}, stride={self.stride}, padding={self.padding}, edge_correction={self.edge_correction}, d={self.d})'

    def initialize_parameters(self, x):
        if is_lazy(self.filter):
            assert len(x.shape) == 2 + self.d
            channels = x.shape[1]
            filter_ = make_gaussian_filter(
                    self.sigma,
                    self.kernel_size,
                    self.d,
                    channels,
            )

            self.filter.materialize(shape=filter_.shape, dtype=filter_.dtype)
            self.filter.copy_(filter_)

        if self.edge_correction and is_lazy(self.weights):
            shape = 1, 1, *x.shape[2:]
            ones = torch.ones(shape, dtype=x.dtype, device=x.device)
            weights = self.blur(ones)

            self.weights.materialize(shape=weights.shape, dtype=weights.dtype)
            self.weights.copy_(weights)

    def forward(self, x):
        y = self.blur(x)

        if self.edge_correction:
            y /= self.weights

        return y

    def blur(self, x):
        return self.conv(
                x,
                self.filter,
                stride=self.stride,
                padding=self.padding,
                groups=x.shape[1],
        )

def make_gaussian_filter(sigma, kernel_size, d, channels):
    grid = torch.meshgrid(
            *[torch.arange(kernel_size)] * d,
            indexing='ij',
    )
    grid = torch.stack(grid, dim=-1)

    mean = (kernel_size - 1) / 2.
    variance = sigma ** 2.

    # setting the dtype is needed, otherwise it becomes an integer tensor
    r = torch.sum((grid - mean) ** 2., dim=-1, dtype=torch.get_default_dtype())

    # Build the gaussian kernel
    _filter = torch.exp(-r / (2 * variance))

    # Normalize
    _filter /= torch.sum(_filter)

    # The filter needs to be reshaped to be used in depthwise convolution
    _filter = _filter\
            .view(1, 1, *[kernel_size]*d)\
            .repeat((channels, 1, *[1]*d))

    return _filter


def kernel_size_from_radius(radius):
    return 2 * int(round(radius)) + 1



from typing import Iterable

def unique_ever_seen(iterable: Iterable) -> Iterable:
    already_seen = set()

    for item in iterable:
        if item in already_seen:
            continue
        else:
            already_seen.add(item)
            yield item






from escnn.group import *
from escnn.gspaces import *
from escnn.nn import FourierFieldType, GeometricTensor

from ..equivariant_module import EquivariantModule
from ..fourier import InverseFourierTransform, FourierTransform

import torch
import torch.nn.functional as F

from typing import List, Tuple, Any, Optional

import numpy as np

__all__ = ["FourierPointwise", "FourierELU"]


class FourierPointwise(EquivariantModule):
    
    def __init__(self, in_type, *args, **kwargs):
        r"""
        
        Perform an Inverse Fourier Transform to sample the input features, apply the pointwise non-linearity in the
        group domain (Dirac-delta basis) and, finally, compute the Fourier Transform to obtain irreps coefficients.
        
        .. warning::
            This operation is only *approximately* equivariant and its equivariance depends on the sampling grid and the
            non-linear activation used, as well as the original band-limitation of the input features.
            
        The same function is applied to every channel independently.
        By default, the input representation is preserved by this operation and, therefore, it equals the output
        representation.
        Optionally, the output can have a different band-limit by using the argument ``out_type``.
        
        The class first constructs a band-limited regular representation of ```gspace.fibergroup``` using
        :meth:`escnn.group.Group.spectral_regular_representation`.
        The band-limitation of this representation is specified by ```irreps``` which should be a list containing
        a list of ids identifying irreps of ```gspace.fibergroup``` (see :attr:`escnn.group.IrreducibleRepresentation.id`).
        This representation is used to define the input and output field types, each containing ```channels``` copies
        of a feature field transforming according to this representation.
        A feature vector transforming according to such representation is interpreted as a vector of coefficients
        parameterizing a function over the group using a band-limited Fourier basis.

        .. note::
            Instead of building the list ``irreps`` manually, most groups implement a method ``bl_irreps()`` which can be
            used to generate this list with through a simpler interface. Check each group's documentation.
        
        To approximate the Fourier transform, this module uses a finite number of samples from the group.
        The set of samples used is specified by the ``grid`` argument.  This can be any list of group elements, but most
        often the :meth:`~escnn.group.Group.grid` is used to generate this list.
        
        Args:
            gspace (GSpace):  the gspace describing the symmetries of the data. The Fourier transform is
                              performed over the group ```gspace.fibergroup```
            channels (int): number of independent fields in the input `FieldType`
            irreps (list): list of irreps' ids to construct the band-limited representation
            function (str): the identifier of the non-linearity.
                    It is used to specify which function to apply.
                    By default (``'p_relu'``), ReLU is used.
            *grid_args: parameters used to construct the discretization grid
            inplace (bool): applies the non-linear activation in-place. Default: `True`
            out_irreps (list, optional): optionally, one can specify a different band-limiting in output
            normalize (bool, optional): if ``True``, the rows of the IFT matrix (and the columns of the FT matrix) are normalized. Default: ``True``
            **grid_kwargs: keyword parameters used to construct the discretization grid
            
        """
        super().__init__()

        if isinstance(in_type, FourierFieldType):
            self._init(in_type, *args, **kwargs)
        else:
            self._init_deprecated(in_type, *args, **kwargs)

    def _init(
            self,
            in_type: FourierFieldType,
            grid: List[GroupElement],
            function: str = 'p_relu',
            *,
            out_type: Optional[FourierFieldType] = None,
            inplace: bool = True,
            normalize: bool = True,
    ):
        self.in_type = in_type
        self.out_type = out_type or in_type

        self.ift = InverseFourierTransform(in_type, grid, normalize=normalize)
        self.ft = FourierTransform(grid, self.out_type, extra_irreps=in_type.bl_irreps, normalize=normalize)

        if callable(function):
            self._function = function
        elif function == 'p_relu':
            self._function = F.relu_ if inplace else F.relu
        elif function == 'p_elu':
            self._function = F.elu_ if inplace else F.elu
        elif function == 'p_sigmoid':
            self._function = torch.sigmoid_ if inplace else F.sigmoid
        elif function == 'p_tanh':
            self._function = torch.tanh_ if inplace else F.tanh
        else:
            raise ValueError('Function "{}" not recognized!'.format(function))
        
    def _init_deprecated(
            self,
            gspace: GSpace,
            channels: int,
            irreps: List,
            *grid_args,
            grid: List[GroupElement] = None,
            function: str = 'p_relu',
            inplace: bool = True,
            out_irreps: List = None,
            normalize: bool = True,
            **grid_kwargs
    ):
        from warnings import warn
        warn("The `gspace`, `channels`, and `irreps` arguments are deprecated.  Use the `in_type` argument instead.", DeprecationWarning)

        in_type = FourierFieldType(gspace, channels, irreps)
        out_type = FourierFieldType(gspace, channels, out_irreps or irreps)

        if grid is None:
            grid = gspace.fibergroup.grid(*grid_args, **grid_kwargs)

        self._init(
                in_type, grid, function,
                out_type=out_type,
                inplace=inplace,
                normalize=normalize,
        )
        
    def forward(self, x_hat: GeometricTensor) -> GeometricTensor:
        r"""

        Applies the pointwise activation function on the input fields

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map after the non-linearities have been applied

        """

        assert x_hat.type == self.in_type

        x = self.ift(x_hat)

        x.tensor = self._function(x.tensor)

        return self.ft(x)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-5, rtol: float = 2e-2, assert_raise: bool = True) -> List[Tuple[Any, float]]:
    
        gspace = self.in_type.gspace
        rho = self.in_type.rho
        rho_out = self.out_type.rho
        c = self.in_type.size
        B = 128
        x = torch.randn(B, c, *[3]*gspace.dimensionality)

        # since we mostly use non-linearities like relu or elu, we make sure 
        # the average value of the features is
        # positive, such that, when we test inputs with only frequency 0 (or only low frequencies), the output is not
        # zero everywhere
        x = x.view(B, len(self.in_type), rho.size, *[3]*gspace.dimensionality)
        p = 0
        for irr in rho.irreps:
            irr = gspace.irrep(*irr)
            if irr.is_trivial():
                x[:, :, p] = x[:, :, p].abs()
            p+=irr.size

        x = x.view(B, self.in_type.size, *[3]*gspace.dimensionality)

        errors = []

        # for el in gspace.testing_elements:
        for _ in range(100):
            
            el = gspace.fibergroup.sample()
    
            x1 = GeometricTensor(x.clone(), self.in_type)
            x2 = GeometricTensor(x.clone(), self.in_type).transform_fibers(el)

            out1 = self(x1).transform_fibers(el)
            out2 = self(x2)

            out1 = out1.tensor.view(B, len(self.out_type), rho_out.size, *out1.shape[2:]).detach().numpy()
            out2 = out2.tensor.view(B, len(self.out_type), rho_out.size, *out2.shape[2:]).detach().numpy()

            errs = np.linalg.norm(out1 - out2, axis=2).reshape(-1)
            errs[errs < atol] = 0.
            norm = np.sqrt(np.linalg.norm(out1, axis=2).reshape(-1) * np.linalg.norm(out2, axis=2).reshape(-1))
            
            relerr = errs / norm

            # print(el, errs.max(), errs.mean(), relerr.max(), relerr.min())

            if assert_raise:
                assert relerr.mean()+ relerr.std() < rtol, \
                    'The error found during equivariance check with element "{}" is too high: max = {}, mean = {}, std ={}' \
                        .format(el, relerr.max(), relerr.mean(), relerr.std())

            # errors.append((el, errs.mean()))
            errors.append(relerr)

        # return errors
        return np.concatenate(errors).reshape(-1)

    # The following properties are just for backwards compatibility
    @property
    def space(self):
        return self.in_type.gspace

    @property
    def rho(self):
        return self.in_type.rho
    
    @property
    def rho_out(self):
        return self.out_type.rho
    


class FourierELU(FourierPointwise):
    
    def __init__(self, *args, **kwargs):
        r"""

        Applies a Inverse Fourier Transform to sample the input features, apply ELU point-wise in the
        group domain (Dirac-delta basis) and, finally, computes the Fourier Transform to obtain irreps coefficients.
        See :class:`~escnn.nn.FourierPointwise` for more details.

        Args:
            gspace (GSpace):  the gspace describing the symmetries of the data. The Fourier transform is
                              performed over the group ```gspace.fibergroup```
            channels (int): number of independent fields in the input `FieldType`
            irreps (list): list of irreps' ids to construct the band-limited representation
            *grid_args: parameters used to construct the discretization grid
            inplace (bool): applies the non-linear activation in-place. Default: `True`
            out_irreps (list, optional): optionally, one can specify a different band-limiting in output
            normalize (bool, optional): if ``True``, the rows of the IFT matrix (and the columns of the FT matrix) are normalized. Default: ``True``
            **grid_kwargs: keyword parameters used to construct the discretization grid

        """
        
        kwargs['function'] = 'p_elu'
        super().__init__(*args, **kwargs)





from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule
from .utils import calc_pool_output_shape, check_dimensions, get_nd_tuple

import torch
import torch.nn.functional as F

from typing import List, Tuple, Any, Union

from math import prod


__all__ = ["NormMaxPool"]

_MAX_POOLS = {
        2: F.max_pool2d,
        3: F.max_pool3d,
}

class _NormMaxPoolND(EquivariantModule):
    
    def __init__(self,
                 *,
                 d: int,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False
                 ):
        r"""
        
        Max-pooling based on the fields' norms. In a given window of shape :attr:`kernel_size`, for each
        group of channels belonging to the same field, the field with the highest norm (as the length of the vector)
        is preserved.
        
        Except :attr:`in_type`, the other parameters correspond to the ones of :class:`torch.nn.MaxPool2d`.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            
        """

        super().__init__()

        check_dimensions(in_type, d)

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        self.kernel_size = get_nd_tuple(kernel_size, d)
        self.stride = get_nd_tuple(stride if stride is not None else kernel_size, d)
        self.padding = get_nd_tuple(padding, d)
        self.dilation = get_nd_tuple(dilation, d)
        self.ceil_mode = ceil_mode
        
        # For each representation size, an "index" which will select all 
        # channels belonging to said representations.  Each "index" could 
        # either by a list of (integer) index numbers, or a slice.
        self._indices = {}
        
        # For each representation size, whether all the representations of that 
        # size are located immediately adjacent to each other.
        _contiguous = {}

        last_stop = 0
        last_size = None

        for r in self.in_type.representations:
            if r.size != last_size:
                _contiguous[r.size] = r.size not in _contiguous
        
            start, stop = last_stop, last_stop + r.size
            self._indices.setdefault(r.size, []).extend(range(start, stop))

            last_size = r.size
            last_stop = stop

        # Use slices for contiguous fields.  Presumably this is a bit more 
        # efficient?
        for k, indices in self._indices.items():
            if _contiguous[k]:
                self._indices[k] = slice(min(indices), max(indices) + 1)
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Run the norm-based max-pooling on the input tensor
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type
        
        b, c, *xyzi = input.tensor.shape
        
        assert len(xyzi) == self.d
        
        b, c, *xyzo = self.evaluate_output_shape(input.tensor.shape)

        # compute the squares of the values of each channel
        input_squared = input.tensor ** 2
        
        # pre-allocate the output tensor
        output = torch.empty(b, c, *xyzo, device=input.tensor.device)
        
        # reshape the input to merge the spatial dimensions
        input = input.tensor.reshape(b, c, -1)
        
        # iterate through all field sizes
        for s, indices in self._indices.items():

            # compute the norms
            norms = input_squared[:, indices, :, :] \
                .view(b, -1, s, *xyzi) \
                .sum(dim=2)
            
            # run max-pooling on the norms-tensor
            _, indx = _MAX_POOLS[self.d](
                    norms,
                    self.kernel_size,
                    self.stride,
                    self.padding,
                    self.dilation,
                    self.ceil_mode,
                    return_indices=True,
            )
            
            # in order to use the pooling indices computed for the norms to retrieve the fields, they need to be
            # expanded in the inner field dimension
            indx = indx.view(b, -1, 1, prod(xyzo)).expand(-1, -1, s, -1)

            # retrieve the fields from the input tensor using the pooling indeces
            output[:, indices, :, :] = input[:, indices, :] \
                .view(b, -1, s, prod(xyzi)) \
                .gather(3, indx) \
                .view(b, -1, *xyzo)
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        return calc_pool_output_shape(self.d, self, input_shape, self.out_type)
        
    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # this kind of pooling is not really equivariant so we can not test 
        # equivariance
        pass


class NormMaxPool2D(_NormMaxPoolND):
    
    def __init__(
            self,
            in_type: FieldType,
            kernel_size: Union[int, Tuple[int, int]],
            stride: Union[int, Tuple[int, int]] = None,
            padding: Union[int, Tuple[int, int]] = 0,
            dilation: Union[int, Tuple[int, int]] = 1,
            ceil_mode: bool = False
    ):
        super().__init__(
                d=2,
                in_type=in_type,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                ceil_mode=ceil_mode,
        )

class NormMaxPool3D(_NormMaxPoolND):
    
    def __init__(
            self,
            in_type: FieldType,
            kernel_size: Union[int, Tuple[int, int]],
            stride: Union[int, Tuple[int, int]] = None,
            padding: Union[int, Tuple[int, int]] = 0,
            dilation: Union[int, Tuple[int, int]] = 1,
            ceil_mode: bool = False
    ):
        super().__init__(
                d=3,
                in_type=in_type,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                ceil_mode=ceil_mode,
        )

# for backward compatibility
NormMaxPool = NormMaxPool2D




from escnn.nn import FieldType

from .pointwise import _PointwisePoolND, _PointwiseMaxPoolAntialiasedND

import torch

from typing import Tuple, Union, Optional

__all__ = [
    "PointwiseMaxPool2D", "PointwiseMaxPool",
    "PointwiseMaxPool3D",
    "PointwiseMaxPoolAntialiased2D", "PointwiseMaxPoolAntialiased",
    "PointwiseMaxPoolAntialiased3D",
]

class PointwiseMaxPool2D(_PointwisePoolND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.MaxPool2d`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.
        
        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when True, will use ceil instead of floor to compute the output shape

        """
        super().__init__(
            in_type, 2,
            pool=torch.nn.MaxPool2d(
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                ceil_mode=ceil_mode,
            ),
        )


class PointwiseMaxPool3D(_PointwisePoolND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int, int]],
                 stride: Union[int, Tuple[int, int, int]] = None,
                 padding: Union[int, Tuple[int, int, int]] = 0,
                 dilation: Union[int, Tuple[int, int, int]] = 1,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.MaxPool3d`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when True, will use ceil instead of floor to compute the output shape

        """
        super().__init__(
            in_type, 3,
            pool=torch.nn.MaxPool3d(
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                ceil_mode=ceil_mode,
            ),
        )


class PointwiseAdaptiveMaxPool2D(_PointwisePoolND):

    def __init__(self,
                 in_type: FieldType,
                 output_size: Union[int, Tuple[int, int]]
                 ):
        r"""

        Module that implements adaptive channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AdaptiveMaxPool2d`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            output_size: the target output size of the image of the form H x W

        """
        super().__init__(
            in_type, 2,
            pool=torch.nn.AdaptiveMaxPool2d(
                output_size=output_size,
            ),
        )


class PointwiseAdaptiveMaxPool3D(_PointwisePoolND):

    def __init__(self,
                 in_type: FieldType,
                 output_size: Union[int, Tuple[int, int, int]]
                 ):
        r"""

        Module that implements adaptive channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AdaptiveMaxPool3d`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            output_size: the target output size of the image of the form H x W

        """
        super().__init__(
            in_type, 3,
            pool=torch.nn.AdaptiveMaxPool3d(
                output_size=output_size,
            ),
        )

class PointwiseMaxPoolAntialiased2D(_PointwiseMaxPoolAntialiasedND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Optional[Union[int, Tuple[int, int]]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 ceil_mode: bool = False,
                 sigma: float = 0.6,
                 ):
        r"""

        Anti-aliased version of channel-wise max-pooling (each channel is treated independently).

        The max over a neighborhood is performed pointwise without downsampling.
        Then, convolution with a Gaussian blurring filter is performed before downsampling the feature map.

        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.


        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            sigma (float): standard deviation for the Gaussian blur filter

        """
        super().__init__(
            in_type, 2,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            ceil_mode=ceil_mode,
            sigma=sigma,
        )


class PointwiseMaxPoolAntialiased3D(_PointwiseMaxPoolAntialiasedND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int, int]],
                 stride: Optional[Union[int, Tuple[int, int, int]]] = None,
                 padding: Union[int, Tuple[int, int, int]] = 0,
                 ceil_mode: bool = False,
                 sigma: float = 0.6,
                 ):
        r"""

        Anti-aliased version of channel-wise max-pooling (each channel is treated independently).

        The max over a neighborhood is performed pointwise withot downsampling.
        Then, convolution with a gaussian blurring filter is performed before downsampling the feature map.

        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.


        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            sigma (float): standard deviation for the Gaussian blur filter

        """
        super().__init__(
            in_type, 3,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            ceil_mode=ceil_mode,
            sigma=sigma,
        )


# for backward compatibility
PointwiseMaxPool = PointwiseMaxPool2D
PointwiseAdaptiveMaxPool = PointwiseAdaptiveMaxPool2D
PointwiseMaxPoolAntialiased = PointwiseMaxPoolAntialiased2D





# first line: 52
@cache.cache
def _clebsh_gordan_tensor(J: Tuple, l: Tuple, j: Tuple, group_class: str, **group_keys) -> np.ndarray:
    
    G = escnn.group.groups_dict[group_class]._generator(**group_keys)
    
    psi_J = G.irrep(*J)
    psi_l = G.irrep(*l)
    psi_j = G.irrep(*j)
    
    D = psi_J.size * psi_l.size * psi_j.size
    
    def build_matrices(samples):
        D_Jl = []
        D_j = []
        for g in samples:
            D_J_g = psi_J(g)
            D_l_g = psi_l(g)
            D_j_g = psi_j(g)
        
            D_Jl_g = np.kron(D_J_g, D_l_g)
            
            D_j.append(D_j_g)
            D_Jl.append(D_Jl_g)
        return D_Jl, D_j
    
    try:
        generators = G.generators
        S = len(generators)
    except ValueError:
        generators = []
        # number of samples to use to approximate the solutions
        # usually 3 are sufficient
        S = 3

    _S = S
    
    while True:
        # sometimes it might not converge, so we need to try a few times
        attepts = 5
        while True:
            try:
                samples = generators + [G.sample() for _ in range(S - len(generators))]
                if len(samples) == 0:
                    basis = np.eye(D)
                else:
                    D_Jl, D_j = build_matrices(samples)
                    basis = find_intertwiner_basis_sylvester(D_Jl, D_j)
                
            except np.linalg.LinAlgError:
                if attepts > 0:
                    attepts -= 1
                    continue
                else:
                    raise
            else:
                break
                
        # check that the solutions found are also in the kernel of the constraint matrix built with other random samples
        D_Jl, D_j = build_matrices(generators + [G.sample() for _ in range(20)])
        tensor = build_sylvester_constraint(D_Jl, D_j).todense().reshape(-1, D)
        
        if np.allclose(tensor @ basis, 0.):
            break
        elif S < MAX_SAMPLES:
            # if this not the case, try again using more samples to build the constraint matrix
            S += 1
        else:
            raise UnderconstrainedCGSystem(G, psi_J.id, psi_l.id, psi_j.id, S)
    
    if S > _S:
        print(G.name, psi_J.id, psi_l.id, psi_j.id, S)

    # the dimensionality of this basis corresponds to the multiplicity of `j` in the tensor-product `J x l`
    s = basis.shape[1]
    assert s % psi_j.sum_of_squares_constituents == 0

    jJl = s // psi_j.sum_of_squares_constituents

    CG = basis.reshape((psi_j.size, psi_J.size, psi_l.size, s)).transpose(1, 2, 3, 0)
    # CG indexed as [J, l, s, j]

    if s == 0:
        return CG

    norm = np.sqrt((CG**2).mean(2, keepdims=True).sum(1, keepdims=True).sum(0, keepdims=True))
    CG /= norm
    
    ortho = np.einsum(
        'Jlsj,Jlti,kji->stk',
        CG, CG, psi_j.endomorphism_basis()
    )
    
    ortho = (ortho**2).sum(2) > 1e-9
    assert ortho.astype(np.uint).sum() == s * psi_j.sum_of_squares_constituents, (ortho, s, jJl, psi_j.sum_of_squares_constituents)

    n, dependencies = connected_components(csgraph=csr_matrix(ortho), directed=False, return_labels=True)
    assert n * psi_j.sum_of_squares_constituents == s, (ortho, n, s, psi_j.sum_of_squares_constituents)

    mask = np.zeros((ortho.shape[0]), dtype=bool)
    for i in range(n):
        columns = np.nonzero(dependencies == i)[0]
        assert len(columns) == psi_j.sum_of_squares_constituents
        selected_column = columns[0]
        mask[selected_column] = 1

    assert mask.sum() == n

    CG = CG[..., mask, :]

    assert CG.shape[-2] == jJl

    B = CG.reshape(-1, psi_j.size * jJl)
    assert np.allclose(B.T@B, np.eye(psi_j.size * jJl))
    
    return CG



# first line: 664
@cache.cache(ignore=['ico'])
def _build_ico_irrep_picklable(ico: Icosahedral, l: int) -> List[Tuple]:
    # To enable caching, the output of this method needs to be picklable so we can not return a dictionary with
    # group elements as keys

    if l == 3:
        
        # Representation of the generator of the cyclic subgroup of order 5
        rho_p = np.zeros((3, 3))
        
        rho_p[0, 0] = rho_p[1, 1] = np.cos(144 / 180. * np.pi)
        rho_p[1, 0] = np.sin(144 / 180. * np.pi)
        rho_p[0, 1] = -np.sin(144 / 180. * np.pi)
        rho_p[2, 2] = 1.
        
        # Representation of the generator of the cyclic subgroup of order 2
        rho_q = np.zeros((3, 3))
        rho_q[0, 0] = 1. / np.sqrt(5)
        rho_q[0, 2] = - 2. / np.sqrt(5)
        rho_q[1, 1] = - 1
        rho_q[2, 0] = - 2. / np.sqrt(5)
        rho_q[2, 2] = - 1. / np.sqrt(5)
        
    elif l == 4:

        # Representation of the generator of the cyclic subgroup of order 5
        rho_p = np.zeros((4, 4))

        rho_p[0, 0] = rho_p[1, 1] = np.cos(72 / 180. * np.pi)
        rho_p[1, 0] = np.sin(72 / 180. * np.pi)
        rho_p[0, 1] = -np.sin(72 / 180. * np.pi)

        rho_p[2, 2] = rho_p[3, 3] = np.cos(144 / 180. * np.pi)
        rho_p[3, 2] = np.sin(144 / 180. * np.pi)
        rho_p[2, 3] = -np.sin(144 / 180. * np.pi)

        # Representation of the generator of the cyclic subgroup of order 2
        rho_q = np.zeros((4, 4))
        rho_q[0, 2] = -1
        rho_q[1, 1] = 2. / np.sqrt(5)
        rho_q[1, 3] = 1. / np.sqrt(5)
        rho_q[2, 0] = -1
        rho_q[3, 1] = 1. / np.sqrt(5)
        rho_q[3, 3] = -2. / np.sqrt(5)

    else:
        raise ValueError()

    generators = [
        (ico._generators[0], rho_p),
        (ico._generators[1], rho_q),
    ]
    
    irreps = generate_irrep_matrices_from_generators(ico, generators)
    return [
        (k.value, k.param, v)
        for k, v in irreps.items()
    ]




from escnn.group import *
from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.nn import init

from ..equivariant_module import EquivariantModule

import torch
from escnn.nn.modules.basismanager import BasisManager
from escnn.nn.modules.basismanager import BlocksBasisExpansion

from torch.nn import Parameter

from typing import List, Tuple, Any

import numpy as np

__all__ = ['TensorProductModule']


class TensorProductModule(EquivariantModule):

    def __init__(self, in_type: FieldType, out_type: FieldType, initialize: bool = True):
        r"""

        Applies a (learnable) quadratic non-linearity to the input features.

        The module requires its input and output types to be *uniform*, i.e. contain multiple copies of the same
        representation; see also :meth:`~escnn.nn.FieldType.uniform`.
        Moreover, the input and output field types must have the same number of fields, i.e.
        ```len(in_type) == len(out_type)```.

        The module computes the tensor product of each field with itself to generate an intermediate feature map.
        Note that this feature map will have size ```len(in_type) * in_type.representations[0].size**2```.
        To prevent the exponential growth of the model's width at each layer, the module includes also a learnable
        linear projection of each ```in_type.representations[0].size**2```-dimensional output field to a corresponding
        ```out_type.representations[0].size``` output field.
        Note that this layer applies an independent linear projection to each field individually but does not mix them.

        ..warning ::
            A model employing only this kind of non-linearities will effectively be a polynomial function.
            Moreover, the degree of the polynomial grows exponentially with the depth of the network.
            This may result in some instabilities during training.

        Args:
            in_type (FieldType): the type of the input field, specifying its transformation law
            out_type (FieldType): the type of the output field, specifying its transformation law
            initialize (bool, optional): initialize the weights of the model (warning: can be slow). Default: ``True``

        Attributes:
            ~.weights (torch.Tensor): the learnable parameters which are used to expand the projection matrix
            ~.matrix (torch.Tensor): the matrix obtained by expanding the parameters in :attr:`~escnn.nn.TensorProductModule.weights`

        """
        super(TensorProductModule, self).__init__()

        self.in_type = in_type
        self.out_type = out_type

        assert self.in_type.gspace == self.out_type.gspace

        self.gspace: GSpace = self.in_type.gspace

        assert in_type.uniform
        assert out_type.uniform
        assert len(in_type) == len(out_type), (len(in_type), len(out_type))

        # number of independent fields
        self._C = len(in_type)

        in_repr = in_type.representations[0]
        hid_repr = in_repr.tensor(in_repr)

        self.hidden_type = self.gspace.type(*[hid_repr]*self._C)

        # BlocksBasisExpansion: submodule which takes care of building the linear projection matrix
        self._basisexpansion = BlocksBasisExpansion(
            # we use only the first representation in input since we want a "grouped" linear layer,not a dense one
            [hid_repr], out_type.representations,
            self.gspace.build_fiber_intertwiner_basis,
            np.zeros((1, 1)),
            recompute=False
        )

        if self.basisexpansion.dimension() == 0:
            raise ValueError('''
                The basis for the steerable matrix is empty!
            ''')

        self.weights = Parameter(torch.zeros(self.basisexpansion.dimension()), requires_grad=True)
        self.weights.data[:] = torch.randn_like(self.weights)
        if initialize:
            # by default, the weights are initialized with a generalized form of He's weight initialization
            init.generalized_he_init(self.weights.data, self.basisexpansion)

        matrix_size = (out_type.size, hid_repr.size)
        self.register_buffer("matrix", torch.zeros(*matrix_size))

    @classmethod
    def construct(cls, gspace: GSpace, channels: int, in_repr: Representation, out_repr: Representation, initialize: bool = True) -> 'TensorProductModule':
        f'''
            Constructor method which provides an alternative way to instantiate a TensorProductModule.
            This method automatically builds the ```in_type``` (respectively, ```out_type```) as ```channels``` copies 
            of ```in_repr``` (```out_repr```).
        '''
        assert in_repr.group == gspace.fibergroup
        assert out_repr.group == gspace.fibergroup
        in_type = gspace.type(*[in_repr]*channels)
        out_type = gspace.type(*[out_repr]*channels)
        return TensorProductModule(in_type, out_type, initialize=initialize)

    def forward(self, input: GeometricTensor):

        assert input.type == self.in_type

        coords = input.coords
        spatial_shape = input.shape[2:]
        input = input.tensor.view(input.shape[0], self._C, self.in_type.representations[0].size,  *spatial_shape)

        # compute tensor product
        tensor_features = torch.einsum('bci...,bco...->bcio...', input, input)
        tensor_features = tensor_features.view(input.shape[0], self._C, self.hidden_type.representations[0].size, *spatial_shape)

        # perform projection
        if not self.training:
            _matrix = self.matrix
        else:
            # retrieve the matrix and the bias
            _matrix = self.expand_parameters()

        _matrix = _matrix.view(self._C, self.out_type.representations[0].size, self.hidden_type.representations[0].size)
        output = torch.einsum('coi,bci...->bco...', _matrix, tensor_features)

        output = output.view(input.shape[0], self.out_type.size, *spatial_shape)

        return self.out_type(output, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        if self.in_type != self.out_type:
            return input_shape[:1] + (self.out_type.size, ) + input_shape[2:]
        else:
            return input_shape

    def export(self):
        raise NotImplementedError()

    @property
    def basisexpansion(self) -> BasisManager:
        r"""
        Submodule which takes care of building the matrix.

        It uses the learnt ``weights`` to expand a basis and returns a matrix in the usual form used by conventional
        convolutional modules.
        It uses the learned ``weights`` to expand the kernel in the G-steerable basis and returns it in the shape
        :math:`(c_\text{out}, c_\text{in}, s^d)`, where :math:`s` is the ``kernel_size`` and :math:`d` is the
        dimensionality of the base space.

        """
        return self._basisexpansion

    def expand_parameters(self) -> torch.Tensor:
        r"""

        Expand the matrix in terms of the :attr:`escnn.nn.TensorProductModule.weights`.

        Returns:
            the expanded projection matrix

        """
        _matrix = self.basisexpansion(self.weights)
        return _matrix.view(self.out_type.size, self.hidden_type.representations[0].size)

    def train(self, mode=True):
        r"""

        If ``mode=True``, the method sets the module in training mode and discards the
        :attr:`~escnn.nn.TensorProductModule.matrix` attribute.

        If ``mode=False``, it sets the module in evaluation mode. Moreover, the method builds the matrix
        using the current values of the trainable parameters and store them in :attr:`~escnn.nn.TensorProductModule.matrix`
        such that it is not recomputed at each forward pass.

        .. warning ::

            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of this class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.

        Args:
            mode (bool, optional): whether to set training mode (``True``) or evaluation mode (``False``).
                                   Default: ``True``.

        """

        if not mode:
            _matrix = self.expand_parameters()
            self.register_buffer("matrix", _matrix)
        else:
            if hasattr(self, "matrix"):
                del self.matrix

        return super(TensorProductModule, self).train(mode)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-4, assert_raise: bool = True) -> List[
        Tuple[Any, float]]:

        c = self.in_type.size
        B = 128
        x = torch.randn(B, c, *[3] * self.gspace.dimensionality)

        errors = []

        # for el in self.gspace.testing_elements:
        for _ in range(100):

            el = self.gspace.fibergroup.sample()

            x1 = GeometricTensor(x.clone(), self.in_type)
            x2 = GeometricTensor(x.clone(), self.in_type).transform_fibers(el)

            out1 = self(x1).transform_fibers(el)
            out2 = self(x2)

            out1 = out1.tensor.view(B, len(self.out_type), self.out_type.representations[0].size, *out1.shape[2:]).detach().numpy()
            out2 = out2.tensor.view(B, len(self.out_type), self.out_type.representations[0].size, *out2.shape[2:]).detach().numpy()

            errs = np.linalg.norm(out1 - out2, axis=2).reshape(-1)
            errs[errs < atol] = 0.
            norm = np.sqrt(np.linalg.norm(out1, axis=2).reshape(-1) * np.linalg.norm(out2, axis=2).reshape(-1))

            relerr = errs / norm

            # print(el, errs.max(), errs.mean(), relerr.max(), relerr.min())

            if assert_raise:
                assert relerr.mean() + relerr.std() < rtol, \
                    'The error found during equivariance check with element "{}" is too high: max = {}, mean = {}, std ={}' \
                        .format(el, relerr.max(), relerr.mean(), relerr.std())

            # errors.append((el, errs.mean()))
            errors.append(relerr)

        # return errors
        return np.concatenate(errors).reshape(-1)
